# coding: utf8

from typing import Optional

from ._forward_curve_types import Steps, Turns
from .._object_definition import ObjectDefinition

from ._enums import (
    DayCountBasis,
    InterpolationMode,
    CalendarAdjustment,
    PriceSide,
    CompoundingType,
    ExtrapolationMode,
)
from ._models import (
    ConvexityAdjustment,
    Step,
    Turn,
)
from ..._types import Strings, OptBool, OptStr
from ...._tools import create_repr


class SwapZcCurveParameters(ObjectDefinition):
    """
    Parameters
    ----------
    interest_calculation_method : InterestCalculationMethod, optional
        Day count basis of the calculated zero coupon rates.
        Default value is: Dcb_Actual_Actual
    calendar_adjustment : CalendarAdjustment, optional
        Cash flow adjustment according to a calendar
        No:
        Null:
        Weekend: for cash flow pricing using the calendar
        WeekendCalendar: for cash flow pricing using the calendar defined
                         by the parameter 'calendars'.
    calendars : string, optional
        A list of one or more calendar codes used to define non-working days and to
        adjust coupon dates and values.
    compounding_type : CompoundingType, optional

    convexity_adjustment : ConvexityAdjustment, optional

    extrapolation_mode : ExtrapolationMode, optional
        None: no extrapolation
        Constant: constant extrapolation
        Linear: linear extrapolation
    interpolation_mode : InterpolationMode, optional
        Interpolation method used in swap zero curve bootstrap.
        Default value is: CubicSpline

        CubicDiscount: local cubic interpolation of discount factors
        CubicRate: local cubic interpolation of rates
        CubicSpline: a natural cubic spline
        Linear: linear interpolation
        Log: log linear interpolation
        ForwardMonotoneConvex
    price_side : SwapPriceSide, optional
        Defines which data is used for the rate surface computation.
        Default value is: Mid
    steps : Step, optional
        Use to calculate the swap rate surface discount curve, when OIS is selected as
        discount curve.
        The steps can specify overnight index stepped dates or/and rates.
    turns : Turn, optional
        Used to include end period rates/turns when calculating swap rate surfaces
    ignore_existing_definition : bool, optional

    reference_tenor : str, optional
        Root tenor(s) for the xIbor dependencies
    use_convexity_adjustment : bool, optional
        false / true
        Default value is: true.
        It indicates if the system needs to retrieve the convexity adjustment
    use_multi_dimensional_solver : bool, optional
        false / true
        Default value is: true.
        Specifies the use of the multi-dimensional solver for yield curve bootstrapping.
        This solving method is required because the bootstrapping method sometimes
        creates a ZC curve which does not accurately reprice the input instruments used
        to build it.
        The multi-dimensional solver is recommended when cubic interpolation methods
        are used in building the curve
        (in other cases, performance might be inferior to the regular bootstrapping
        method).
         - true: to use multi-dimensional solver for yield curve bootstrapping
         - false: not to use multi-dimensional solver for yield curve bootstrapping
    use_steps : bool, optional
        false / true
        Default value is: false.
        It indicates if the system needs to retrieve the overnight index
        stepped dates or/and rates
    valuation_date : str, optional
        The valuation date
        Default value is the current date
    """

    def __init__(
        self,
        interest_calculation_method: Optional[DayCountBasis] = None,
        calendar_adjustment: Optional[CalendarAdjustment] = None,
        calendars: Strings = None,
        compounding_type: Optional[CompoundingType] = None,
        convexity_adjustment: Optional[ConvexityAdjustment] = None,
        extrapolation_mode: Optional[ExtrapolationMode] = None,
        interpolation_mode: Optional[InterpolationMode] = None,
        price_side: Optional[PriceSide] = None,
        steps: Steps = None,
        turns: Turns = None,
        ignore_existing_definition: OptBool = None,
        reference_tenor: OptStr = None,
        use_convexity_adjustment: OptBool = None,
        use_multi_dimensional_solver: OptBool = None,
        use_steps: OptBool = None,
        valuation_date: OptStr = None,
    ) -> None:
        super().__init__()
        self.interest_calculation_method = interest_calculation_method
        self.calendar_adjustment = calendar_adjustment
        self.calendars = calendars
        self.compounding_type = compounding_type
        self.convexity_adjustment = convexity_adjustment
        self.extrapolation_mode = extrapolation_mode
        self.interpolation_mode = interpolation_mode
        self.price_side = price_side
        self.steps = steps
        self.turns = turns
        self.ignore_existing_definition = ignore_existing_definition
        self.reference_tenor = reference_tenor
        self.use_convexity_adjustment = use_convexity_adjustment
        self.use_multi_dimensional_solver = use_multi_dimensional_solver
        self.use_steps = use_steps
        self.valuation_date = valuation_date

    def __repr__(self):
        return create_repr(
            self,
            middle_path="curves.forward_curve",
            class_name=self.__class__.__name__,
        )

    @property
    def calendar_adjustment(self):
        """
        Cash flow adjustment according to a calendar
        No:
        Null:
        Weekend: for cash flow pricing using the calendar
        WeekendCalendar: for cash flow pricing using the calendar defined
                         by the parameter 'calendars'.
        :return: enum CalendarAdjustment
        """
        return self._get_enum_parameter(CalendarAdjustment, "calendarAdjustment")

    @calendar_adjustment.setter
    def calendar_adjustment(self, value):
        self._set_enum_parameter(CalendarAdjustment, "calendarAdjustment", value)

    @property
    def calendars(self):
        """
        A list of one or more calendar codes used to define non-working days and to
        adjust coupon dates and values.
        :return: list string
        """
        return self._get_list_parameter(str, "calendars")

    @calendars.setter
    def calendars(self, value):
        self._set_list_parameter(str, "calendars", value)

    @property
    def compounding_type(self):
        """
        :return: enum CompoundingType
        """
        return self._get_enum_parameter(CompoundingType, "compoundingType")

    @compounding_type.setter
    def compounding_type(self, value):
        self._set_enum_parameter(CompoundingType, "compoundingType", value)

    @property
    def convexity_adjustment(self):
        """
        :return: object ConvexityAdjustment
        """
        return self._get_object_parameter(ConvexityAdjustment, "convexityAdjustment")

    @convexity_adjustment.setter
    def convexity_adjustment(self, value):
        self._set_object_parameter(ConvexityAdjustment, "convexityAdjustment", value)

    @property
    def extrapolation_mode(self):
        """
        None: no extrapolation
        Constant: constant extrapolation
        Linear: linear extrapolation
        :return: enum ExtrapolationMode
        """
        return self._get_enum_parameter(ExtrapolationMode, "extrapolationMode")

    @extrapolation_mode.setter
    def extrapolation_mode(self, value):
        self._set_enum_parameter(ExtrapolationMode, "extrapolationMode", value)

    @property
    def interest_calculation_method(self):
        """
        Day count basis of the calculated zero coupon rates.
        Default value is: Dcb_Actual_Actual
        :return: enum DayCountBasis
        """
        return self._get_enum_parameter(DayCountBasis, "interestCalculationMethod")

    @interest_calculation_method.setter
    def interest_calculation_method(self, value):
        self._set_enum_parameter(DayCountBasis, "interestCalculationMethod", value)

    @property
    def interpolation_mode(self):
        """
        Interpolation method used in swap zero curve bootstrap.
        Default value is: CubicSpline

        CubicDiscount: local cubic interpolation of discount factors
        CubicRate: local cubic interpolation of rates
        CubicSpline: a natural cubic spline
        Linear: linear interpolation
        Log: log linear interpolation
        ForwardMonotoneConvex
        :return: enum InterpolationMode
        """
        return self._get_enum_parameter(InterpolationMode, "interpolationMode")

    @interpolation_mode.setter
    def interpolation_mode(self, value):
        self._set_enum_parameter(InterpolationMode, "interpolationMode", value)

    @property
    def price_side(self):
        """
        Defines which data is used for the rate surface computation.
        Default value is: Mid
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def steps(self):
        """
        Use to calculate the swap rate surface discount curve, when OIS is selected as
        discount curve.
        The steps can specify overnight index stepped dates or/and rates.
        :return: list Step
        """
        return self._get_list_parameter(Step, "steps")

    @steps.setter
    def steps(self, value):
        self._set_list_parameter(Step, "steps", value)

    @property
    def turns(self):
        """
        Used to include end period rates/turns when calculating swap rate surfaces
        :return: list Turn
        """
        return self._get_list_parameter(Turn, "turns")

    @turns.setter
    def turns(self, value):
        self._set_list_parameter(Turn, "turns", value)

    @property
    def ignore_existing_definition(self):
        """
        :return: bool
        """
        return self._get_parameter("ignoreExistingDefinition")

    @ignore_existing_definition.setter
    def ignore_existing_definition(self, value):
        self._set_parameter("ignoreExistingDefinition", value)

    @property
    def reference_tenor(self):
        """
        Root tenor(s) for the xIbor dependencies
        :return: str
        """
        return self._get_parameter("referenceTenor")

    @reference_tenor.setter
    def reference_tenor(self, value):
        self._set_parameter("referenceTenor", value)

    @property
    def use_convexity_adjustment(self):
        """
        false / true
        Default value is: true.
        It indicates if the system needs to retrieve the convexity adjustment
        :return: bool
        """
        return self._get_parameter("useConvexityAdjustment")

    @use_convexity_adjustment.setter
    def use_convexity_adjustment(self, value):
        self._set_parameter("useConvexityAdjustment", value)

    @property
    def use_multi_dimensional_solver(self):
        """
        false / true
        Default value is: true.
        Specifies the use of the multi-dimensional solver for yield curve bootstrapping.
        This solving method is required because the bootstrapping method sometimes
        creates a ZC curve which does not accurately reprice the input instruments used
        to build it.
        The multi-dimensional solver is recommended when cubic interpolation methods
        are used in building the curve
        (in other cases, performance might be inferior to the regular bootstrapping
        method).
         - true: to use multi-dimensional solver for yield curve bootstrapping
         - false: not to use multi-dimensional solver for yield curve bootstrapping
        :return: bool
        """
        return self._get_parameter("useMultiDimensionalSolver")

    @use_multi_dimensional_solver.setter
    def use_multi_dimensional_solver(self, value):
        self._set_parameter("useMultiDimensionalSolver", value)

    @property
    def use_steps(self):
        """
        false / true
        Default value is: false.
        It indicates if the system needs to retrieve the overnight index
        stepped dates or/and rates
        :return: bool
        """
        return self._get_parameter("useSteps")

    @use_steps.setter
    def use_steps(self, value):
        self._set_parameter("useSteps", value)

    @property
    def valuation_date(self):
        """
        The valuation date
        Default value is the current date
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



# coding: utf8


from ._enums import VolatilityAdjustmentType
from ._enums import Axis
from ._enums import InputVolatilityType
from .._object_definition import ObjectDefinition


class IIrVolModelPricingParameters(ObjectDefinition):
    def __init__(
        self,
        input_volatility_type=None,
        volatility_adjustment_type=None,
        x_axis=None,
        y_axis=None,
        z_axis=None,
        market_data_date=None,
        shift_percent=None,
        source=None,
        stripping_shift_percent=None,
        valuation_date=None,
    ):
        super().__init__()
        self.input_volatility_type = input_volatility_type
        self.volatility_adjustment_type = volatility_adjustment_type
        self.x_axis = x_axis
        self.y_axis = y_axis
        self.z_axis = z_axis
        self.market_data_date = market_data_date
        self.shift_percent = shift_percent
        self.source = source
        self.stripping_shift_percent = stripping_shift_percent
        self.valuation_date = valuation_date

    @property
    def input_volatility_type(self):
        """
        :return: enum InputVolatilityType
        """
        return self._get_enum_parameter(InputVolatilityType, "inputVolatilityType")

    @input_volatility_type.setter
    def input_volatility_type(self, value):
        self._set_enum_parameter(InputVolatilityType, "inputVolatilityType", value)

    @property
    def volatility_adjustment_type(self):
        """
        Volatility Adjustment method for stripping: ConstantCaplet, ConstantCap, ShiftedCap, NormalizedCap, NormalizedCaplet
        :return: enum VolatilityAdjustmentType
        """
        return self._get_enum_parameter(
            VolatilityAdjustmentType, "volatilityAdjustmentType"
        )

    @volatility_adjustment_type.setter
    def volatility_adjustment_type(self, value):
        self._set_enum_parameter(
            VolatilityAdjustmentType, "volatilityAdjustmentType", value
        )

    @property
    def x_axis(self):
        """
        Specifies the unit for the x axis (e.g. Date, Tenor)
        :return: enum Axis
        """
        return self._get_enum_parameter(Axis, "xAxis")

    @x_axis.setter
    def x_axis(self, value):
        self._set_enum_parameter(Axis, "xAxis", value)

    @property
    def y_axis(self):
        """
        Specifies the unit for the y axis (e.g. Strike, Delta). This may depend on the asset class.
        For Fx Volatility Surface, we support both Delta and Strike.
        :return: enum Axis
        """
        return self._get_enum_parameter(Axis, "yAxis")

    @y_axis.setter
    def y_axis(self, value):
        self._set_enum_parameter(Axis, "yAxis", value)

    @property
    def z_axis(self):
        """
        Specifies the unit for the z axis (e.g. Strike, Tenor, Expiries). This applies on Ir SABR Volatility Cube.
        :return: enum Axis
        """
        return self._get_enum_parameter(Axis, "zAxis")

    @z_axis.setter
    def z_axis(self, value):
        self._set_enum_parameter(Axis, "zAxis", value)

    @property
    def market_data_date(self):
        """
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def shift_percent(self):
        """
        Shift value to use in calibration(Strike/Forward). Default: 0.0
        :return: float
        """
        return self._get_parameter("shiftPercent")

    @shift_percent.setter
    def shift_percent(self, value):
        self._set_parameter("shiftPercent", value)

    @property
    def source(self):
        """
        Requested volatility data contributor.
        :return: str
        """
        return self._get_parameter("source")

    @source.setter
    def source(self, value):
        self._set_parameter("source", value)

    @property
    def stripping_shift_percent(self):
        """
        Shift value to use in caplets stripping(Strike/Forward). Default: 0.0
        :return: float
        """
        return self._get_parameter("strippingShiftPercent")

    @stripping_shift_percent.setter
    def stripping_shift_percent(self, value):
        self._set_parameter("strippingShiftPercent", value)

    @property
    def valuation_date(self):
        """
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



import copy
from typing import Any, Callable, List, TYPE_CHECKING

from ._stream import StreamingChain
from ..._types import ExtendedParams, OptBool, OptInt, OptStr
from ...._core.session import get_valid_session
from ...._tools import cached_property, create_repr
from ....delivery._stream.base_stream import StreamOpenWithUpdatesMixin

if TYPE_CHECKING:
    from ...._core.session import Session


class Stream(StreamOpenWithUpdatesMixin):
    """
    Stream is designed to request streaming chains and decode it dynamically.
    This class also act like a cache for each part of the chain record.

    Parameters
    ----------
    name : str
        Single instrument name
    session : Session, optional
        The Session defines the source where you want to retrieve your data
    service : str, optional
        Name service
    skip_summary_links : bool, optional
        Store skip summary links
    skip_empty : bool, optional
        Store skip empty
    override_summary_links : int, optional
        Store the override number of summary links
    extended_params : dict, optional
        If necessary other parameters

    Methods
    -------
    open(**kwargs)
        Open the Stream connection

    close()
        Closes the Stream connection, releases resources

    is_chain
        True - stream was decoded as a chain
        False - stream wasn't identified as a chain

    Attributes
    __________
    constituents: list
        A list of constituents in the chain record or empty list

    """

    def __init__(
        self,
        name: str,
        session: "Session" = None,
        service: OptStr = None,
        skip_summary_links: OptBool = True,
        skip_empty: OptBool = True,
        override_summary_links: OptInt = None,
        extended_params: ExtendedParams = None,
    ) -> None:
        session = get_valid_session(session)
        self._name = name
        self._session = session
        self._service = service
        self._skip_summary_links = skip_summary_links
        self._skip_empty = skip_empty
        self._override_summary_links = override_summary_links
        self._extended_params = extended_params

    @cached_property
    def _stream(self) -> StreamingChain:
        streaming_chain = StreamingChain(
            name=self._name,
            session=self._session,
            service=self._service,
            skip_summary_links=self._skip_summary_links,
            skip_empty=self._skip_empty,
            override_summary_links=self._override_summary_links,
            extended_params=self._extended_params,
        )
        return streaming_chain

    @property
    def name(self) -> str:
        return self._stream.name

    @property
    def is_chain(self) -> bool:
        return self._stream.is_chain

    @property
    def num_summary_links(self) -> int:
        return self._stream.num_summary_links

    @property
    def summary_links(self) -> List[str]:
        return self._stream.summary_links

    @property
    def display_name(self) -> str:
        return self._stream.display_name

    @property
    def constituents(self) -> List[str]:
        return copy.deepcopy(self._stream.get_constituents())

    def on_add(self, func: Callable[[int, str, "Stream"], Any]) -> "Stream":
        func = make_callback(self, func)
        self._stream.on_add(func)
        return self

    def on_remove(self, func: Callable[[str, int, "Stream"], Any]) -> "Stream":
        func = make_callback(self, func)
        self._stream.on_remove(func)
        return self

    def on_update(self, func: Callable[[str, str, int, "Stream"], Any]) -> "Stream":
        func = make_callback(self, func)
        self._stream.on_update(func)
        return self

    def on_complete(self, func: Callable[[list, "Stream"], Any]) -> "Stream":
        func = make_callback(self, func)
        self._stream.on_complete(func)
        return self

    def on_error(self, func: Callable[[str, tuple, "Stream"], Any]) -> "Stream":
        func = make_error_callback(self, func)
        self._stream.on_error(func)
        return self

    def __repr__(self):
        return create_repr(
            self,
            content=f"{{name='{self._name}'}}",
            class_name=self.__class__.__name__,
        )


def make_callback(stream: Stream, func: Callable) -> Callable:
    """Return a callback functions with correct arguments order."""

    def callback(*args):
        func(*args, stream)

    return callback


def make_error_callback(stream: Stream, func: Callable) -> Callable:
    """Return a callback function with correct arguments order for error handling."""

    def callback(*args):
        args = reversed(args)
        func(*args, stream)

    return callback



from typing import Any


def create_repr(
    obj: object,
    middle_path: str = "",
    class_name: str = "Definition",
    content: Any = None,
) -> str:
    module_path = obj.__module__
    new_path = get_new_path(module_path, middle_path, class_name)
    obj_id = hex(id(obj))
    s = f"<{new_path} object at {obj_id}>"

    if content:
        s = s.replace(">", f" {content}>")

    return s


def get_new_path(module_path: str, middle: str = "", class_name: str = "") -> str:
    """
    Examples
    --------
    >>> module_path = "refinitiv.data.pricing.stream"
    >>> class_name = "Definition"
    >>> start = "refinitiv.data"
    >>> middle = "stream"
    >>> new_path = ".".join([start, middle, class_name])
    >>> new_path
    ... "refinitiv.data.stream.Definition"

    >>> middle = "content"
    >>> new_path
    ... "refinitiv.data.content.Definition"
    """
    start, *_ = module_path.split("._", maxsplit=1)

    if middle:
        new_path = ".".join([start, middle, class_name])
    else:
        new_path = ".".join([start, class_name])
    return new_path



__all__ = (
    "annual",
    "historical_snapshots_non_periodic_measures",
    "historical_snapshots_periodic_measures_annual",
    "historical_snapshots_periodic_measures_interim",
    "historical_snapshots_recommendations",
    "interim",
    "non_periodic_measures",
    "recommendations",
)

from . import (
    annual,
    historical_snapshots_non_periodic_measures,
    historical_snapshots_periodic_measures_annual,
    historical_snapshots_periodic_measures_interim,
    historical_snapshots_recommendations,
    interim,
    non_periodic_measures,
    recommendations,
)



__all__ = (
    "BermudanSwaptionDefinition",
    "BuySell",
    "PricingParameters",
    "CallPut",
    "Definition",
    "ExerciseScheduleType",
    "ExerciseStyle",
    "SwaptionMarketDataRule",
    "SwaptionSettlementType",
)

from ._bermudan_swaption_definition import BermudanSwaptionDefinition
from ._definition import Definition
from ._enums import (
    ExerciseScheduleType,
    BuySell,
    CallPut,
    ExerciseStyle,
    SwaptionSettlementType,
)
from ._swaption_market_data_rule import SwaptionMarketDataRule
from ._swaption_pricing_parameters import PricingParameters



# coding: utf8

from enum import Enum, unique

from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class ForwardCurvesOutputs(Enum):
    CONSTITUENTS = "Constituents"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            ForwardCurvesOutputs, _FORWARD_CURVES_OUTPUTS_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _FORWARD_CURVES_OUTPUTS_VALUES_IN_LOWER_BY_FORWARD_CURVES_OUTPUTS, some
        )


_FORWARD_CURVES_OUTPUTS_VALUES = tuple(t.value for t in ForwardCurvesOutputs)
_FORWARD_CURVES_OUTPUTS_VALUES_IN_LOWER_BY_FORWARD_CURVES_OUTPUTS = {
    name.lower(): item for name, item in ForwardCurvesOutputs.__members__.items()
}



# coding: utf8
# contract_gen 2020-05-18 08:30:59.205820

__all__ = ["InterestType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class InterestType(Enum):
    """
    A flag that indicates whether the leg is fixed or float.

    Possible values are:
        Fixed - the leg has a fixed coupon.
        Float - the leg has a floating rate index. Mandatory.
    """

    FIXED = "Fixed"
    FLOAT = "Float"
    STEPPED = "Stepped"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(InterestType, _INTEREST_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_INTEREST_TYPE_VALUES_IN_LOWER_BY_INTEREST_TYPE, some)


_INTEREST_TYPE_VALUES = tuple(t.value for t in InterestType)
_INTEREST_TYPE_VALUES_IN_LOWER_BY_INTEREST_TYPE = {
    name.lower(): item for name, item in InterestType.__members__.items()
}



# coding: utf8

from typing import Optional

from ..._instrument_definition import ObjectDefinition


class FxDualCurrencyDefinition(ObjectDefinition):
    """
    Parameters
    ----------
    deposit_start_date : str, optional
        Deposit Start Date for the DualCurrencyDeposit option
    margin_percent : float, optional
        Margin for the DualCurrencyDeposit option
    """

    def __init__(
        self,
        deposit_start_date: Optional[str] = None,
        margin_percent: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.deposit_start_date = deposit_start_date
        self.margin_percent = margin_percent

    @property
    def deposit_start_date(self):
        """
        Deposit Start Date for the DualCurrencyDeposit option
        :return: str
        """
        return self._get_parameter("depositStartDate")

    @deposit_start_date.setter
    def deposit_start_date(self, value):
        self._set_parameter("depositStartDate", value)

    @property
    def margin_percent(self):
        """
        Margin for the DualCurrencyDeposit option
        :return: float
        """
        return self._get_parameter("marginPercent")

    @margin_percent.setter
    def margin_percent(self, value):
        self._set_parameter("marginPercent", value)



# coding: utf8

from typing import Optional

from .._instrument_pricing_parameters import InstrumentPricingParameters
from ._enums import PriceSide


class PricingParameters(InstrumentPricingParameters):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    price_side : PriceSide, optional
        Price Side to consider when retrieving Market Data.
    income_tax_percent : float, optional
        Income tax percent is subtracted from applied interest rate percents at the end
        of the deposit. Example: "5" means 5%
        By default is 0.
    market_data_date : str, optional
        The market data date for pricing.
        By default, the market_data_date date is the valuation_date or Today.
    valuation_date : str, optional
        The valuation date for pricing. If not set the valuation date is equal to
        market_data_date or Today. For assets that contains a settlementConvention, the
        default valuation date  is equal to the settlementdate of the Asset that is
        usually the TradeDate+SettlementConvention.

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> rdf.term_deposit.PricingParameters(valuation_date="2020-04-24")
    """

    def __init__(
        self,
        price_side: Optional[PriceSide] = None,
        income_tax_percent: Optional[float] = None,
        market_data_date: Optional[str] = None,
        valuation_date: Optional[str] = None,
    ):
        super().__init__()
        self.price_side = price_side
        self.income_tax_percent = income_tax_percent
        self.market_data_date = market_data_date
        self.valuation_date = valuation_date

    @property
    def price_side(self):
        """
        Price Side to consider when retrieving Market Data.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def income_tax_percent(self):
        """
        Income tax percent is substracted from applied interest rate percents in the end of deposit.
        Example: "5" means 5%
        :return: float
        """
        return self._get_parameter("incomeTaxPercent")

    @income_tax_percent.setter
    def income_tax_percent(self, value):
        self._set_parameter("incomeTaxPercent", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.
        If not set the valuation date is equal to MarketDataDate or Today. For assets that contains a settlementConvention,
        the default valuation date  is equal to the settlementdate of the Asset that is usually the TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing.
        By default, the marketDataDate date is the ValuationDate or Today.
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)



import urllib

from pandas import DataFrame
from refinitiv.data._tools import urljoin

from ._data_types import BucketData, FileData, FileSetData, PackageData
from ._iter_object import IterObj
from ._tools import _get_query_params
from .._data._data_provider import (
    DataProvider,
    ResponseFactory,
    RequestFactory,
)


# --------------------------------------------------------------------------------------
#   Request factory
# --------------------------------------------------------------------------------------


class CFSRequestFactory(RequestFactory):
    def get_query_parameters(self, *_, **kwargs) -> list:
        return _get_query_params(**kwargs)

    def add_query_parameters(self, url, query_parameters) -> str:
        return "?".join([url, urllib.parse.urlencode(query_parameters, safe=";")])


class CFSPackageRequestFactory(CFSRequestFactory):
    def get_query_parameters(self, *_, **kwargs) -> list:
        package_id = kwargs.get("_package_id")
        if package_id is not None:
            return []

        return super().get_query_parameters(**kwargs)

    def get_url(self, *args, **kwargs):
        url = super().get_url(*args, **kwargs)
        url_id = kwargs.get("_package_id")
        if url_id is not None:
            url = urljoin(url, url_id)
        return url


class CFSStreamRequestFactory(RequestFactory):
    def get_url(self, *args, **kwargs):
        return super().get_url(*args, **kwargs) + "/{id}/stream"

    def get_path_parameters(self, *_, **kwargs) -> dict:
        path_parameters = kwargs.get("path_parameters") or dict()
        file_id = kwargs.get("id")
        if file_id:
            path_parameters["id"] = file_id
        return path_parameters

    def get_query_parameters(self, *_, **kwargs) -> list:
        query_parameters = kwargs.get("query_parameters") or []
        query_parameters.append(("doNotRedirect", "true"))
        return query_parameters


# --------------------------------------------------------------------------------------
#   Response factory
# --------------------------------------------------------------------------------------


def cfs_build_df(raw, **kwargs):
    _value = raw.get("value") or [raw]
    _columns = set()
    for i in _value:
        _columns = _columns | i.keys()
    _columns = tuple(_columns)
    _data = [
        [value[key] if key in value else None for key in _columns] for value in _value
    ]
    _dataframe = DataFrame(_data, columns=_columns)
    return _dataframe


class CFSResponseFactory(ResponseFactory):
    def create_success(self, data, *args, **kwargs):
        session = args[0]
        inst = self.response_class(is_success=True, **data)
        dfbuilder = self.get_dfbuilder(kwargs)
        content_data = data.get("content_data")
        content_value = content_data.get("value") or [content_data]
        _iter_obj = IterObj(content_value, session, self.data_class)
        inst.data = self.data_class(content_data, _iter_obj, dfbuilder=dfbuilder)
        inst.data._owner = inst
        return inst

    def create_fail(self, data, *args, **kwargs):
        inst = self.response_class(is_success=False, **data)
        inst.data = self.data_class(data)
        inst.data._owner = inst
        return inst


# --------------------------------------------------------------------------------------
#   Data provider
# --------------------------------------------------------------------------------------


class CFSDataProvider(DataProvider):
    def __init__(self, data_class, request=CFSRequestFactory()):
        super().__init__(
            response=CFSResponseFactory(data_class=data_class),
            request=request,
        )


cfs_buckets_data_provider = CFSDataProvider(data_class=BucketData)
cfs_file_sets_data_provider = CFSDataProvider(data_class=FileSetData)
cfs_files_data_provider = CFSDataProvider(data_class=FileData)
cfs_packages_data_provider = CFSDataProvider(
    data_class=PackageData, request=CFSPackageRequestFactory()
)
cfs_stream_data_provider = DataProvider(request=CFSStreamRequestFactory())



from typing import Union, Optional, TYPE_CHECKING, Iterable

from ._symbol_type import SYMBOL_TYPE_VALUES
from ._symbology_provider_layer import SymbologyContentProviderLayer
from ..._tools import create_repr


if TYPE_CHECKING:
    from ._asset_class import AssetClass
    from ._country_code import CountryCode
    from ._asset_state import AssetState
    from ._symbol_type import SymbolTypes
    from .._types import ExtendedParams, OptStr, Strings


class Definition(SymbologyContentProviderLayer):
    """
    This class describe parameters to retrieve data for symbol conversion.

    Parameters
    ----------
    symbols: str or list of str
        Single instrument or list of instruments to convert.

    from_symbol_type: SymbolTypes, optional
        Instrument code to convert from.
        Possible values: 'CUSIP', 'ISIN', 'SEDOL', 'RIC', 'ticker', 'lipperID', 'IMO'
        Default: '_AllUnique'

    to_symbol_types: SymbolTypes, str or list of str or SymbolTypes, optional
        Instrument code to convert to.
        Possible values: 'CUSIP', 'ISIN', 'SEDOL', 'RIC', 'ticker', 'lipperID', 'IMO', 'OAPermID'
        Default: all symbol types are requested

    closure: str, optional
        Specifies the parameter that will be merged with the request

    extended_params: dict, optional
        Other parameters can be provided if necessary

    preferred_country_code: CountryCode, optional
        Unique ISO 3166 code for country

    asset_class: AssetClass, optional
        AssetClass value to build filter parameter.

    asset_state: AssetState, optional
        AssetState value to build filter parameter.

    Examples
    --------
    >>> from refinitiv.data.content import symbol_conversion
    >>> definition = symbol_conversion.Definition(
    >>>     symbols=["US5949181045", "US02079K1079"],
    >>>     from_symbol_type=symbol_conversion.SymbolTypes.ISIN,
    >>>     to_symbol_types=[symbol_conversion.SymbolTypes.RIC, symbol_conversion.SymbolTypes.OA_PERM_ID],
    >>>     preferred_country_code=symbol_conversion.CountryCode.USA,
    >>>     asset_class=[
    >>>        symbol_conversion.AssetClass.COMMODITIES,
    >>>        symbol_conversion.AssetClass.EQUITIES,
    >>>        symbol_conversion.AssetClass.WARRANTS
    >>>     ],
    >>>     asset_state=symbol_conversion.AssetState.INACTIVE)
    """

    def __init__(
        self,
        symbols: Union["Strings", str],
        from_symbol_type: Optional["SymbolTypes"] = "_AllUnique",
        to_symbol_types: Union[Iterable, str, "SymbolTypes"] = SYMBOL_TYPE_VALUES,
        closure: "OptStr" = None,
        extended_params: "ExtendedParams" = None,
        preferred_country_code: Optional["CountryCode"] = None,
        asset_class: Union[Iterable, "AssetClass"] = None,
        asset_state: Optional["AssetState"] = None,
    ):
        self._symbols = symbols
        self._from_symbol_type = from_symbol_type
        self._to_symbol_types = to_symbol_types
        self._closure = closure
        self._extended_params = extended_params
        self.preferred_country_code = preferred_country_code
        self.asset_class = asset_class
        self.asset_state = asset_state

        super().__init__(
            symbols=self._symbols,
            from_symbol_type=self._from_symbol_type,
            to_symbol_types=self._to_symbol_types,
            closure=self._closure,
            extended_params=self._extended_params,
            preferred_country_code=self.preferred_country_code,
            asset_class=self.asset_class,
            asset_state=self.asset_state,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="content.symbols_convention",
            content=f"{{symbols='{self._symbols}'}}",
        )



from typing import List

import pandas as pd
from pandas.core.dtypes.common import is_datetime64_any_dtype


def get_name_columns_without_datetime(df: pd.DataFrame) -> List[str]:
    return [
        column_name
        for column_name, dtype in df.dtypes.to_dict().items()
        if not is_datetime64_any_dtype(dtype)
    ]



# coding: utf8

__all__ = ["FxPoint"]

from .._object_definition import ObjectDefinition
from .._enums import Status


class FxPoint(ObjectDefinition):
    def __init__(
        self,
        bid=None,
        ask=None,
        mid=None,
        status=None,
        instrument=None,
        processing_information=None,
        spot_decimals=None,
    ):
        super().__init__()
        self.bid = bid
        self.ask = ask
        self.mid = mid
        self.status = status
        self.instrument = instrument
        self.processing_information = processing_information
        self.spot_decimals = spot_decimals

    @property
    def status(self):
        """
        :return: enum Status
        """
        return self._get_enum_parameter(Status, "status")

    @status.setter
    def status(self, value):
        self._set_enum_parameter(Status, "status", value)

    @property
    def ask(self):
        """
        :return: float
        """
        return self._get_parameter("ask")

    @ask.setter
    def ask(self, value):
        self._set_parameter("ask", value)

    @property
    def bid(self):
        """
        :return: float
        """
        return self._get_parameter("bid")

    @bid.setter
    def bid(self, value):
        self._set_parameter("bid", value)

    @property
    def instrument(self):
        """
        :return: str
        """
        return self._get_parameter("instrument")

    @instrument.setter
    def instrument(self, value):
        self._set_parameter("instrument", value)

    @property
    def mid(self):
        """
        :return: float
        """
        return self._get_parameter("mid")

    @mid.setter
    def mid(self, value):
        self._set_parameter("mid", value)

    @property
    def processing_information(self):
        """
        :return: str
        """
        return self._get_parameter("processingInformation")

    @processing_information.setter
    def processing_information(self, value):
        self._set_parameter("processingInformation", value)

    @property
    def spot_decimals(self):
        """
        :return: str
        """
        return self._get_parameter("spotDecimals")

    @spot_decimals.setter
    def spot_decimals(self, value):
        self._set_parameter("spotDecimals", value)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.226816

__all__ = ["FxSwapCalculationMethod"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class FxSwapCalculationMethod(Enum):
    DEPOSIT_CCY1_IMPLIED_FROM_FX_SWAP = "DepositCcy1ImpliedFromFxSwap"
    DEPOSIT_CCY2_IMPLIED_FROM_FX_SWAP = "DepositCcy2ImpliedFromFxSwap"
    FX_SWAP = "FxSwap"
    FX_SWAP_IMPLIED_FROM_DEPOSIT = "FxSwapImpliedFromDeposit"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            FxSwapCalculationMethod, _FX_SWAP_CALCULATION_METHOD_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _FX_SWAP_CALCULATION_METHOD_VALUES_IN_LOWER_BY_FX_SWAP_CALCULATION_METHOD,
            some,
        )


_FX_SWAP_CALCULATION_METHOD_VALUES = tuple(t.value for t in FxSwapCalculationMethod)
_FX_SWAP_CALCULATION_METHOD_VALUES_IN_LOWER_BY_FX_SWAP_CALCULATION_METHOD = {
    name.lower(): item for name, item in FxSwapCalculationMethod.__members__.items()
}



# coding: utf8

__all__ = ["DiscountingType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class DiscountingType(Enum):
    LIBOR_DISCOUNTING = "LiborDiscounting"
    OIS_DISCOUNTING = "OisDiscounting"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(DiscountingType, _DISCOUNTING_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_DISCOUNTING_TYPE_VALUES_IN_LOWER_BY_DISCOUNTING_TYPE, some)


_DISCOUNTING_TYPE_VALUES = tuple(t.value for t in DiscountingType)
_DISCOUNTING_TYPE_VALUES_IN_LOWER_BY_DISCOUNTING_TYPE = {
    name.lower(): item for name, item in DiscountingType.__members__.items()
}



__all__ = (
    "Axis",
    "Definition",
    "DiscountingType",
    "Format",
    "InputVolatilityType",
    "SurfaceLayout",
    "SwaptionCalculationParams",
    "SwaptionSurfaceDefinition",
    "VolatilityAdjustmentType",
)

from ._definition import Definition
from ._enums import (
    DiscountingType,
    VolatilityAdjustmentType,
    Axis,
    InputVolatilityType,
    Format,
)
from ._models import SurfaceLayout
from ..._surfaces._i_ir_vol_model_definition import (
    IIrVolModelDefinition as SwaptionSurfaceDefinition,
)
from ..._surfaces._i_ir_vol_model_pricing_parameters import (
    IIrVolModelPricingParameters as SwaptionCalculationParams,
)



# coding: utf8


from .._enums import DoubleBinaryType, SettlementType
from ..._instrument_definition import ObjectDefinition


class FxDoubleBinaryDefinition(ObjectDefinition):
    """ """

    def __init__(
        self,
        double_binary_type=None,
        payout_amount=None,
        payout_ccy=None,
        settlement_type=None,
        trigger_down=None,
        trigger_up=None,
    ):
        super().__init__()
        self.double_binary_type = double_binary_type
        self.payout_amount = payout_amount
        self.payout_ccy = payout_ccy
        self.settlement_type = settlement_type
        self.trigger_down = trigger_down
        self.trigger_up = trigger_up

    @property
    def double_binary_type(self):
        """
        Possible values:
            - None,
            - DoubleNoTouch
        :return: fx.DoubleBinaryType
        """
        return self._get_enum_parameter(DoubleBinaryType, "binaryType")

    @double_binary_type.setter
    def double_binary_type(self, value):
        self._set_enum_parameter(DoubleBinaryType, "binaryType", value)

    @property
    def payout_amount(self):
        """
        :return: double
        """
        return self._get_parameter("payoutAmount")

    @payout_amount.setter
    def payout_amount(self, value):
        self._set_parameter("payoutAmount", value)

    @property
    def payout_ccy(self):
        """
        :return: string
        """
        return self._get_parameter("payoutCcy")

    @payout_ccy.setter
    def payout_ccy(self, value):
        self._set_parameter("payoutCcy", value)

    @property
    def settlement_type(self):
        """
        Settlement Type of the BinaryOption
        Possible values :
            - Undefined
            - Cash
            - Asset
        :return: enum fx.SettlementType
        """
        return self._get_enum_parameter(SettlementType, "settlementType")

    @settlement_type.setter
    def settlement_type(self, value):
        self._set_enum_parameter(SettlementType, "settlementType", value)

    @property
    def trigger_down(self):
        """
        Barrier Down for the Double Barrier option
        :return: double
        """
        return self._get_parameter("triggerDown")

    @trigger_down.setter
    def trigger_down(self, value):
        self._set_parameter("triggerDown", value)

    @property
    def trigger_up(self):
        """
        Barrier Up for the Double Barrier option
        :return: double
        """
        return self._get_parameter("triggerUp")

    @trigger_up.setter
    def trigger_up(self, value):
        self._set_parameter("triggerUp", value)



# coding: utf-8

from urllib.request import getproxies

from ..._tools import parse_url


class ProxyInfo(object):

    __proxies_info = None
    __no_proxy = None

    @classmethod
    def get_proxies_info(cls):
        if cls.__proxies_info is None:
            cls.__proxies_info = {}
            proxies = getproxies()
            for scheme, proxy_data in proxies.items():
                if scheme == "no":
                    cls.__no_proxy = proxy_data.split(",")
                    if not cls.__no_proxy:
                        # if no_proxy is empty, rely on default value (websocket._url.DEFAULT_NO_PROXY_HOST)
                        cls.__no_proxy = None
                else:
                    cls.__proxies_info[scheme] = ProxyInfo.proxy_info_from_url(
                        proxy_data
                    )

        return cls.__proxies_info

    @classmethod
    def get_no_proxy(cls):
        # call cls.get_proxies_info() to be sure class attr were initialized
        cls.get_proxies_info()
        return cls.__no_proxy

    @classmethod
    def get_proxy_info(cls, scheme):
        proxies_info = cls.get_proxies_info()
        return proxies_info[scheme] if scheme in proxies_info else None

    def __init__(
        self,
        proxy_type,
        proxy_host,
        proxy_port,
        proxy_user=None,
        proxy_pass=None,
    ):
        """Args:
        proxy_type: The type of proxy server ('http')
        proxy_host: The hostname or IP address of the proxy server.
        proxy_port: The port that the proxy server is running on.
        proxy_user: The username used to authenticate with the proxy server.
        proxy_pass: The password used to authenticate with the proxy server.
        """
        if isinstance(proxy_user, bytes):
            proxy_user = proxy_user.decode()
        if isinstance(proxy_pass, bytes):
            proxy_pass = proxy_pass.decode()
        (
            self._proxy_type,
            self._proxy_host,
            self._proxy_port,
            self._proxy_user,
            self._proxy_pass,
        ) = (
            proxy_type,
            proxy_host,
            proxy_port,
            proxy_user,
            proxy_pass,
        )

    @property
    def type(self):
        return self._proxy_type

    @property
    def host(self):
        return self._proxy_host

    @property
    def port(self):
        return self._proxy_port

    @property
    def user(self):
        return self._proxy_user

    @property
    def password(self):
        return self._proxy_pass

    @property
    def auth(self):
        if self._proxy_user and self._proxy_pass:
            return tuple(self._proxy_user, self._proxy_pass)

    def __repr__(self):
        return (
            f"<ProxyInfo type={self.type} host={self.host}:{self.port} "
            + f" user={self.user} pass={self.password}>"
        )

    def proxy_json_info(self):
        return {
            "type": self._proxy_type,
            "host": self._proxy_host,
            "port": self._proxy_port,
            "user": self._proxy_user,
            "pass": self._proxy_pass,
        }

    @staticmethod
    def proxy_info_from_url(url):
        """Construct a ProxyInfo from a URL (such as http_proxy env var value)"""
        url = parse_url(url)
        username = None
        password = None
        port = None
        method = url[0]
        if "@" in url[1]:
            ident, host_port = url[1].split("@", 1)
            if ":" in ident:
                username, password = ident.split(":", 1)
            else:
                password = ident
        else:
            host_port = url[1]
        if ":" in host_port:
            host, port = host_port.split(":", 1)
        else:
            host = host_port

        if port:
            port = int(port)
        else:
            try:
                port = dict(https=443, http=80)[method]
            except KeyError:
                pass

        pi = ProxyInfo(
            proxy_type=method,
            proxy_host=host,
            proxy_port=port,
            proxy_user=username or None,
            proxy_pass=password or None,
        )
        return pi



# coding: utf8

__all__ = ["BarrierStyle"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class BarrierStyle(Enum):
    NONE = "None"
    AMERICAN = "American"
    EUROPEAN = "European"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(BarrierStyle, _BARRIER_STYLE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_BARRIER_STYLE_VALUES_IN_LOWER_BY_BARRIER_STYLE, some)


_BARRIER_STYLE_VALUES = tuple(t.value for t in BarrierStyle)
_BARRIER_STYLE_VALUES_IN_LOWER_BY_BARRIER_STYLE = {
    name.lower(): item for name, item in BarrierStyle.__members__.items()
}



# coding: utf8

from typing import Optional

from .._instrument_definition import ObjectDefinition
from ._enums import (
    FxLegType,
    BuySell,
)


class LegDefinition(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    start_date : str, optional

    end_date : str, optional
        The maturity date of the contract that is the date the amounts are exchanged.
        Either the end_date or the tenor must be provided.
    tenor : str, optional
        The tenor representing the maturity date of the contract (e.g. '1Y' or '6M' ).
        Either the end_date or the tenor must be provided.
    leg_tag : str, optional
        A user defined string to identify the leg. Optional.
    deal_ccy_buy_sell : BuySell, optional
        The direction of the trade in terms of the deal currency : 'Buy' or 'Sell'.
        Optional. Defaults to 'Buy'
    fx_leg_type : FxLegType, optional
        The enumeration that specifies the type of the leg : 'Spot', 'FxForward',
        'FxNonDeliverableForward', 'SwapNear' or 'SwapFar'. Mandatory for MultiLeg,
        FwdFwdSwap, or Swap contracts. Optional for Spot and Forwards contracts.
    contra_amount : float, optional
        The unsigned amount exchanged to buy or sell the traded amount. Optional. By
        default, it is calculated from the traded rate and the deal_amount. If no traded
        rate is provided the market rate will be used.
    contra_ccy : str, optional
        The currency that is exchanged. Optional. By default, the second currency in the
        FxCrossCode.
    deal_amount : float, optional
        The unsigned amount of traded currency actually bought or sold. Optional.
        Defaults to 1,000,000'.
    deal_ccy : str, optional
        The ISO code of the traded currency (e.g. 'EUR' ). Optional. Defaults to the
        first currency of the FxCrossCode.
    start_tenor : str, optional
        The tenor representing the Starting of maturity period of the contract (e.g.
        '1Y' or '6M' ). Either the start_date or the start_tenor must be provided for
        TimeOptionForward.
    """

    def __init__(
        self,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        leg_tag: Optional[str] = None,
        deal_ccy_buy_sell: Optional[BuySell] = None,
        fx_leg_type: Optional[FxLegType] = None,
        contra_amount: Optional[float] = None,
        contra_ccy: Optional[str] = None,
        deal_amount: Optional[float] = None,
        deal_ccy: Optional[str] = None,
        start_tenor: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.start_date = start_date
        self.end_date = end_date
        self.tenor = tenor
        self.leg_tag = leg_tag
        self.deal_ccy_buy_sell = deal_ccy_buy_sell
        self.fx_leg_type = fx_leg_type
        self.contra_amount = contra_amount
        self.contra_ccy = contra_ccy
        self.deal_amount = deal_amount
        self.deal_ccy = deal_ccy
        self.start_tenor = start_tenor

    @property
    def deal_ccy_buy_sell(self):
        """
        The direction of the trade in terms of the deal currency : 'Buy' or 'Sell'.
        Optional. Defaults to 'Buy'
        :return: enum BuySell
        """
        return self._get_enum_parameter(BuySell, "dealCcyBuySell")

    @deal_ccy_buy_sell.setter
    def deal_ccy_buy_sell(self, value):
        self._set_enum_parameter(BuySell, "dealCcyBuySell", value)

    @property
    def fx_leg_type(self):
        """
        The enumeration that specifies the type of the leg : 'Spot', 'FxForward',
        'FxNonDeliverableForward', 'SwapNear' or 'SwapFar'. Mandatory for MultiLeg,
        FwdFwdSwap, or Swap contracts. Optional for Spot and Forwards contracts.
        :return: enum FxLegType
        """
        return self._get_enum_parameter(FxLegType, "fxLegType")

    @fx_leg_type.setter
    def fx_leg_type(self, value):
        self._set_enum_parameter(FxLegType, "fxLegType", value)

    @property
    def contra_amount(self):
        """
        The unsigned amount exchanged to buy or sell the traded amount. Optional. By
        default, it is calculated from the traded rate and the deal_amount. If no traded
        rate is provided the market rate will be used.
        :return: float
        """
        return self._get_parameter("contraAmount")

    @contra_amount.setter
    def contra_amount(self, value):
        self._set_parameter("contraAmount", value)

    @property
    def contra_ccy(self):
        """
        The currency that is exchanged. Optional. By default, the second currency in the
        FxCrossCode.
        :return: str
        """
        return self._get_parameter("contraCcy")

    @contra_ccy.setter
    def contra_ccy(self, value):
        self._set_parameter("contraCcy", value)

    @property
    def deal_amount(self):
        """
        The unsigned amount of traded currency actually bought or sold. Optional.
        Defaults to 1,000,000'.
        :return: float
        """
        return self._get_parameter("dealAmount")

    @deal_amount.setter
    def deal_amount(self, value):
        self._set_parameter("dealAmount", value)

    @property
    def deal_ccy(self):
        """
        The ISO code of the traded currency (e.g. 'EUR' ). Optional. Defaults to the
        first currency of the FxCrossCode.
        :return: str
        """
        return self._get_parameter("dealCcy")

    @deal_ccy.setter
    def deal_ccy(self, value):
        self._set_parameter("dealCcy", value)

    @property
    def end_date(self):
        """
        The maturity date of the contract that is the date the amounts are exchanged.
        Either the end_date or the tenor must be provided.
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def leg_tag(self):
        """
        A user defined string to identify the leg. Optional.
        :return: str
        """
        return self._get_parameter("legTag")

    @leg_tag.setter
    def leg_tag(self, value):
        self._set_parameter("legTag", value)

    @property
    def start_date(self):
        """
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        self._set_parameter("startDate", value)

    @property
    def start_tenor(self):
        """
        The tenor representing the Starting of maturity period of the contract (e.g.
        '1Y' or '6M' ). Either the start_date or the Starttenor must be provided for
        TimeOptionForward.
        :return: str
        """
        return self._get_parameter("startTenor")

    @start_tenor.setter
    def start_tenor(self, value):
        self._set_parameter("startTenor", value)

    @property
    def tenor(self):
        """
        The tenor representing the maturity date of the contract (e.g. '1Y' or '6M' ).
        Either the end_date or the tenor must be provided.
        :return: str
        """
        return self._get_parameter("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_parameter("tenor", value)



__all__ = (
    "esg",
    "estimates",
    "filings",
    "fundamental_and_reference",
    "historical_pricing",
    "ipa",
    "news",
    "ownership",
    "pricing",
    "Response",
    "search",
    "symbol_conversion",
    "trade_data_service",
    "custom_instruments",
)

from . import (
    esg,
    estimates,
    filings,
    fundamental_and_reference,
    historical_pricing,
    ipa,
    news,
    ownership,
    pricing,
    search,
    symbol_conversion,
    trade_data_service,
    custom_instruments,
)
from ..delivery._data._data_provider import Response



# coding: utf8

__all__ = ["PricingParameters"]

from .._instrument_pricing_parameters import InstrumentPricingParameters
from ._enums import IndexConvexityAdjustmentMethod
from ._enums import IndexConvexityAdjustmentIntegrationMethod


class PricingParameters(InstrumentPricingParameters):
    def __init__(
        self,
        valuation_date=None,
        report_ccy=None,
        market_data_date=None,
        index_convexity_adjustment_integration_method=None,
        index_convexity_adjustment_method=None,
        discounting_ccy=None,
        discounting_tenor=None,
        market_value_in_deal_ccy=None,
    ):
        super().__init__()
        self.valuation_date = valuation_date
        self.report_ccy = report_ccy
        self.market_data_date = market_data_date
        self.market_value_in_deal_ccy = market_value_in_deal_ccy
        self.index_convexity_adjustment_integration_method = (
            index_convexity_adjustment_integration_method
        )
        self.index_convexity_adjustment_method = index_convexity_adjustment_method
        self.discounting_ccy = discounting_ccy
        self.discounting_tenor = discounting_tenor

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.
        Optional. If not set the valuation date is equal to MarketDataDate or Today. For assets that contains a
        settlementConvention, the default valuation date  is equal to the settlementdate of the Asset that is usually the
        TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)

    @property
    def report_ccy(self):
        """
        :return: str
        """
        return self._get_parameter("reportCcy")

    @report_ccy.setter
    def report_ccy(self, value):
        self._set_parameter("reportCcy", value)

    @property
    def market_data_date(self):
        """
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def market_value_in_deal_ccy(self):
        """
        :return: float
        """
        return self._get_parameter("marketValueInDealCcy")

    @market_value_in_deal_ccy.setter
    def market_value_in_deal_ccy(self, value):
        self._set_parameter("marketValueInDealCcy", value)

    @property
    def discounting_tenor(self):
        """
        :return: str
        """
        return self._get_parameter("discountingTenor")

    @discounting_tenor.setter
    def discounting_tenor(self, value):
        self._set_parameter("discountingTenor", value)

    @property
    def discounting_ccy(self):
        """
        :return: str
        """
        return self._get_parameter("discountingCcy")

    @discounting_ccy.setter
    def discounting_ccy(self, value):
        self._set_parameter("discountingCcy", value)

    @property
    def index_convexity_adjustment_integration_method(self):
        """
        Integration method used for static replication method.
        :return: enum IndexConvexityAdjustmentIntegrationMethod
        """
        return self._get_enum_parameter(
            IndexConvexityAdjustmentIntegrationMethod,
            "indexConvexityAdjustmentIntegrationMethod",
        )

    @index_convexity_adjustment_integration_method.setter
    def index_convexity_adjustment_integration_method(self, value):
        self._set_enum_parameter(
            IndexConvexityAdjustmentIntegrationMethod,
            "indexConvexityAdjustmentIntegrationMethod",
            value,
        )

    @property
    def index_convexity_adjustment_method(self):
        """
        Convexity adjustment type for CMS swaps and Libor in arrears swaps. Values can be: None, BlackScholes, LiborSwapMethod,
        or Replication
        :return: enum IndexConvexityAdjustmentMethod
        """
        return self._get_enum_parameter(
            IndexConvexityAdjustmentMethod, "indexConvexityAdjustmentMethod"
        )

    @index_convexity_adjustment_method.setter
    def index_convexity_adjustment_method(self, value):
        self._set_enum_parameter(
            IndexConvexityAdjustmentMethod, "indexConvexityAdjustmentMethod", value
        )



# coding: utf-8

import json
import os
import queue
import threading
import traceback
from concurrent.futures import Future
from typing import Callable, Dict, List, Optional, TYPE_CHECKING, Union

import websocket
from pyee import EventEmitter

from .event import StreamCxnEvent
from .rwf.socket import RwfSocketApp
from .stream_cxn_state import StreamCxnState
from ..._core.log_reporter import LogReporter
from ..._core.session import SessionType
from ..._core.session.tools import get_delays
from ..._tools import DEBUG

if TYPE_CHECKING:
    from ._stream_cxn_config import StreamCxnConfig
    from ..._core.session import Session

LOGIN_STREAM_ID = 2
MAX_LISTENERS = 2000
WAIT_TIMEOUT_FOR_CLOSE_MESSAGE_FROM_WS_SERVER = 1
delays = get_delays()


class StreamConnection(threading.Thread, LogReporter):
    _listener: Optional[Union[websocket.WebSocketApp, RwfSocketApp]] = None
    _num_reconnect = 0

    def __init__(
        self,
        connection_id: int,
        name: str,
        session: "Session",
        config: "StreamCxnConfig",
        max_reconnect: int,
    ) -> None:
        self._id: int = connection_id
        self._session: "Session" = session
        self._config: "StreamCxnConfig" = config

        LogReporter.__init__(self, logger=session.logger())
        threading.Thread.__init__(
            self, target=self.connect, name=f"Thread{name}", daemon=True
        )

        self._state: StreamCxnState = StreamCxnState.Initial
        self._is_emit_reconnected: bool = False

        self._is_auto_reconnect: bool = self._session.stream_auto_reconnection
        self._max_reconnect: int = max_reconnect

        self._emitter: EventEmitter = EventEmitter()
        self._emitter.max_listeners = MAX_LISTENERS

        self.prepared: Future = Future()
        self.closed: Future = Future()
        self._timer = threading.Event()

        self._msg_queue: Optional[queue.Queue] = None
        self._msg_processor: Optional[threading.Thread] = None

        self._classname = f"[{name}]"

    @property
    def session(self) -> "Session":
        return self._session

    @property
    def id(self) -> int:
        return self._id

    @property
    def state(self) -> StreamCxnState:
        return self._state

    @property
    def subprotocol(self) -> str:
        return ""

    @property
    def can_connect(self) -> bool:
        return self.state in {StreamCxnState.Reconnecting, StreamCxnState.Initial}

    @property
    def is_connecting(self) -> bool:
        return self.state is StreamCxnState.Connecting

    @property
    def is_connected(self) -> bool:
        return self.state is StreamCxnState.Connected

    @property
    def is_ready(self) -> bool:
        return self.state is StreamCxnState.Ready

    @property
    def can_disconnect(self) -> bool:
        return self.is_ready or self.is_connected or self.is_connecting

    @property
    def is_disconnecting(self) -> bool:
        return self.state is StreamCxnState.Disconnecting

    @property
    def is_disposed(self) -> bool:
        return self.state is StreamCxnState.Disposed

    @property
    def can_reconnect(self) -> bool:
        if self.session.server_mode and self._is_auto_reconnect:
            return True

        if self.session.server_mode and not self._is_auto_reconnect:
            return False

        if not self.session.server_mode and self._is_auto_reconnect:
            num_urls = len(self._config.urls)
            if self._num_reconnect >= self._max_reconnect * num_urls:
                return False
            else:
                return True
        else:
            return False

    def connect(self) -> None:
        if self.is_connecting or self.is_connected:
            self.debug(f"{self._classname} can’t connect, state={self.state}")
            return

        self.debug(f"{self._classname} is connecting [con]")

        self._state = StreamCxnState.Connecting
        self._emitter.emit(StreamCxnEvent.CONNECTING, self)

        self._msg_queue = queue.Queue()
        self._msg_processor = threading.Thread(
            target=self._process_messages, name=f"Msg-Proc-{self.name}", daemon=True
        )
        self._msg_processor.start()

        if self._config.transport == "tcp":
            self._run_tcp_listener()
        else:
            self._run_websocket_listener()

    def _run_tcp_listener(self):
        def on_open(app):
            self._on_ws_open(app)
            self._state = StreamCxnState.Ready
            self.prepared.set_result(True)

        host, port = self._config.url.split(":")
        port = int(port) if port else None

        cfg = self._session.config
        cfg_prefix = "apis.streaming.pricing"
        field_dict_path = cfg.get(f"{cfg_prefix}.field_dict_path")
        enumtype_path = cfg.get(f"{cfg_prefix}.enumtype_path")

        self.debug(
            f"{self._classname} connect (\n"
            f"\tnum_connect={self._num_reconnect},\n"
            f"\turl={self._config.url},\n"
            f"\ttransport={self._config.transport},\n"
        )

        self._listener = RwfSocketApp(
            host=host,
            port=port,
            field_dict_path=field_dict_path,
            enumtype_path=enumtype_path,
            on_open=on_open,
            on_msg=self._on_ws_message,
            on_close=self._on_ws_close,
        )

        self._listener.run_forever(self.get_login_message())

    def _run_websocket_listener(self):
        if DEBUG:
            websocket.enableTrace(True)

        cookie = None
        user_id = os.getenv("REFINITIV_AAA_USER_ID")
        if self._session.type == SessionType.DESKTOP and user_id:
            cookie = f"user-uuid={user_id}"

        headers = ["User-Agent: Python"] + self._config.headers
        subprotocols = [self.subprotocol]

        self.debug(
            f"{self._classname} connect (\n"
            f"\tnum_connect={self._num_reconnect},\n"
            f"\turl={self._config.url},\n"
            f"\theaders={headers},\n"
            f"\tcookies={cookie},\n"
            f"\ttransport={self._config.transport},\n"
            f"\tsubprotocols={subprotocols})"
        )

        self._listener = websocket.WebSocketApp(
            url=self._config.url,
            header=headers,
            cookie=cookie,
            on_open=self._on_ws_open,
            on_message=self._on_ws_message,
            on_error=self._on_ws_error,
            on_close=self._on_ws_close,
            on_ping=self._on_ws_ping,
            on_pong=self._on_ws_pong,
            subprotocols=subprotocols,
        )

        proxy_config = self._config.proxy_config
        http_proxy_host = None
        http_proxy_port = None
        http_proxy_auth = None
        proxy_type = None
        if proxy_config:
            http_proxy_host = proxy_config.host
            http_proxy_port = proxy_config.port
            http_proxy_auth = proxy_config.auth
            proxy_type = proxy_config.type

        no_proxy = self._config.no_proxy

        self._listener.run_forever(
            http_proxy_host=http_proxy_host,
            http_proxy_port=http_proxy_port,
            http_proxy_auth=http_proxy_auth,
            http_no_proxy=no_proxy,
            proxy_type=proxy_type,
            skip_utf8_validation=True,
        )

    def disconnect(self) -> None:
        if self.is_disconnecting or self.is_disposed:
            self.debug(f"{self._classname} can’t disconnect, state={self.state}")
            return

        self.debug(f"{self._classname} is disconnecting [dis]")
        self._state = StreamCxnState.Disconnecting
        self.prepared.cancel()
        self._emitter.emit(StreamCxnEvent.DISCONNECTING, self)
        self.debug(f"{self._classname} disconnected [DIS]")

    def dispose(self) -> None:
        if self.is_disposed:
            self.debug(f"{self._classname} can’t dispose, state={self.state}")
            return

        self.debug(f"{self._classname} is disposing [d]")

        close_message = self.get_close_message()
        if close_message and not self.is_connecting:
            self.send(close_message)

        self._state = StreamCxnState.Disposed

        self._listener.close()
        self._listener = None

        self._msg_queue = None
        self._msg_processor = None

        if not self.prepared.done():
            self.prepared.set_result(False)
        self.closed.set_result(True)
        self.prepared.cancel()
        self.closed.cancel()
        self._emitter.emit(StreamCxnEvent.DISPOSED, self)
        self.debug(f"{self._classname} disposed [D]")

    def get_login_message(self) -> dict:
        # for override
        pass

    def get_close_message(self) -> dict:
        # for override
        pass

    def send_login_message(self) -> None:
        login_message = self.get_login_message()
        self.send(login_message)

    def send(self, data: dict) -> None:
        s = json.dumps(data)
        self.debug(f"{self._classname} send s={s}")
        self._listener.send(s)

    def on(self, event: str, listener: Callable) -> None:
        self._emitter.on(event, listener)

    def remove_listener(self, event: str, listener: Callable) -> None:
        self._emitter.remove_listener(event, listener)

    def _on_ws_open(self, ws: websocket.WebSocketApp) -> None:
        self.debug(f"{self._classname} connected [CON]")
        self.debug(f"{self._classname} on_ws_open")

        self._state = StreamCxnState.Connected

        delays.reset()

        self._emitter.emit(StreamCxnEvent.CONNECTED, self)

        login_message = self.get_login_message()
        self.send(login_message)

    def _on_ws_message(self, ws: websocket.WebSocketApp, s: str) -> None:
        self.debug(f"{self._classname} on_ws_message {s}")

        try:
            messages = json.loads(s)
        except UnicodeDecodeError:
            messages = "".join(map(chr, [byte for byte in bytearray(s)]))
            messages = json.loads(messages)

        if self.is_connected:
            if len(messages) > 1:
                raise ValueError(
                    f"Cannot process messages more then one, num={len(messages)}"
                )

            message = messages[0]
            self._handle_login_message(message)

            if self._is_emit_reconnected:
                self.debug(f"Reconnecting is over, emit event Reconnected.")
                self._is_emit_reconnected = False
                self._emitter.emit(StreamCxnEvent.RECONNECTED, self)

        elif self.is_ready or self.is_disconnecting:
            self._msg_queue.put(messages)

        else:
            debug_msg = (
                f"{self._classname} _on_ws_message: "
                f"don't know what to do, {self.state}, "
                f"message={messages}"
            )
            if DEBUG:
                raise ValueError(debug_msg)

            else:
                self.error(debug_msg)

    def _handle_login_message(self, message: dict):
        # for override
        pass

    def _process_messages(self) -> None:
        while not self.is_disposed:
            messages: List[Dict] = self._msg_queue.get()
            for message in messages:
                self._process_message(message)
            self._msg_queue.task_done()

    def _process_message(self, message: dict) -> None:
        # for override
        pass

    def _on_ws_close(
        self, ws: websocket.WebSocketApp, close_status_code: str, close_msg: str
    ) -> None:
        self.debug(
            f"{self._classname} on_ws_close "
            f"(close_status_code={close_status_code}, close_msg={close_msg})"
        )

        if self.is_disposed:
            # do nothing
            pass

        elif self.is_disconnecting:
            self.debug(f"{self._classname} call soon closed.set_result(True)")
            self.closed.set_result(True)

        elif self.is_connecting or self.is_connected or self.is_ready:
            if self.can_reconnect:
                self._state = StreamCxnState.Reconnecting
                self._is_emit_reconnected = True
                if self.prepared.done():
                    self.prepared = Future()
                self._num_reconnect += 1
                self._config.set_next_url()
                delay_secs = delays.next()
                self.debug(
                    f"{self._classname} try to reconnect over url {self._config.url} "
                    f"in {delay_secs} secs, "
                    f"number of reconnections is {self._num_reconnect}"
                )
                self._timer.wait(delay_secs)
                self.connect()

            else:
                self._state = StreamCxnState.Disconnected
                self._emitter.emit(StreamCxnEvent.DISCONNECTED, self)
                self.dispose()

        else:
            debug_msg = (
                f"{self._classname} _on_ws_close: "
                f"don't know what to do, {self.state}"
            )
            if DEBUG:
                raise ValueError(debug_msg)

            else:
                self.error(debug_msg)

    def _on_ws_error(self, ws: websocket.WebSocketApp, exc: Exception) -> None:
        self.debug(f"{self._classname} on_ws_error")

        if DEBUG:
            self.debug(f"{traceback.format_exc()}")

        self.debug(f"{self._classname} Exception: {exc}")

    def _on_ws_ping(self, data: dict) -> None:
        self.debug(f"{self._classname} ping data={data}")

    def _on_ws_pong(self, ws, data: dict) -> None:
        self.debug(f"{self._classname} pong data={data}")

    def __str__(self) -> str:
        s = (
            f"{self.__class__.__name__} {{\n"
            f"\t\tname={self.name},\n"
            f"\t\tsubprotocol={self.subprotocol},\n"
            f"\t\tis_auto_reconnect={self._is_auto_reconnect},\n"
            f"\t\tmax_reconnect={self._max_reconnect},\n"
            f"\t\tcan_reconnect={self.can_reconnect}}}"
        )
        return s



from enum import Enum, unique
from typing import Union

from .._tools import make_enum_arg_parser


class DayIntervalType(Enum):
    INTRA = 0
    INTER = 1


@unique
class Intervals(Enum):
    """
    The list of interval types of the boundary is described below.

    The supported values of intervals :

        Time:

        Backend will return complete N-minute summaries data.
        When the request start and/or end does not at the N minutes boundary,
        the response will be adjusted.

            ONE_MINUTE - return complete 1-minute
            FIVE_MINUTES - return complete 5-minutes
            TEN_MINUTES - return complete 10-minutes
            THIRTY_MINUTES - return complete 30-minutes
            SIXTY_MINUTES - return complete 60-minutes
            ONE_HOUR - return complete 1-hour

        Days:

            DAILY - This is end of day, daily data
            SEVEN_DAYS - Weekly boundary based on the exchange's
                         week summarization definition
            WEEKLY - Weekly boundary based on the exchange's
                     week summarization definition
            MONTHLY - Monthly boundary based on calendar month
            QUARTERLY - Quarterly boundary based on calendar quarter
            TWELVE_MONTHS - Yearly boundary based on calendar year
            YEARLY - Yearly boundary based on calendar year
    """

    ONE_MINUTE = "PT1M"
    FIVE_MINUTES = "PT5M"
    TEN_MINUTES = "PT10M"
    THIRTY_MINUTES = "PT30M"
    SIXTY_MINUTES = "PT60M"
    ONE_HOUR = "PT1H"
    DAILY = "P1D"
    SEVEN_DAYS = "P7D"
    WEEKLY = "P1W"
    MONTHLY = "P1M"
    QUARTERLY = "P3M"
    TWELVE_MONTHS = "P12M"
    YEARLY = "P1Y"


_ISO8601_INTERVALS = [k.value for k in Intervals]
"""['PT1M', 'PT5M', 'PT10M', 'PT30M', 'PT60M', 'PT1H']"""
_INTRADAY = _ISO8601_INTERVALS[:6]
"""['P1D', 'P7D', 'P1W', 'P1M', 'P3M', 'P12M', 'P1Y']"""
_INTERDAY = _ISO8601_INTERVALS[6:]

interval_arg_parser = make_enum_arg_parser(Intervals, can_be_lower=True)


def get_day_interval_type(
    interval: Union[str, Intervals, DayIntervalType]
) -> DayIntervalType:
    if isinstance(interval, DayIntervalType):
        return interval

    interval = interval_arg_parser.get_str(interval)

    if interval in _INTRADAY:
        day_interval_type = DayIntervalType.INTRA

    elif interval in _INTERDAY:
        day_interval_type = DayIntervalType.INTER

    else:
        raise TypeError(f"Incorrect day interval, interval={interval}.")

    return day_interval_type



# coding: utf8
# contract_gen 2020-05-18 08:30:59.244816

__all__ = ["Method"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class Method(Enum):
    AMERICAN_MONTE_CARLO = "AmericanMonteCarlo"
    ANALYTIC = "Analytic"
    MONTE_CARLO = "MonteCarlo"
    PDE = "PDE"
    TREE = "Tree"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(Method, _METHOD_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_METHOD_VALUES_IN_LOWER_BY_METHOD, some)


_METHOD_VALUES = tuple(t.value for t in Method)
_METHOD_VALUES_IN_LOWER_BY_METHOD = {
    name.lower(): item for name, item in Method.__members__.items()
}



from ._data_provider import is_instrument_id, get_correct_symbol
from ._stream import CustomInstrumentsStream
from ...delivery._data._data_provider import DataProviderLayer
from ...delivery._data._endpoint_data import RequestMethod


class CustomInstrumentDataProviderLayer:
    def __init__(self, *args, **kwargs):
        self._data_provider_layer = DataProviderLayer(*args, **kwargs)

    def get_stream(self, session=None) -> CustomInstrumentsStream:
        universe = self._data_provider_layer._kwargs.get("universe")
        if is_instrument_id.match(universe):
            instrument_response = self._data_provider_layer.get_data(
                session, method=RequestMethod.GET
            )
            name = instrument_response.data.raw.get("symbol")
        else:
            name = get_correct_symbol(universe, session)
        stream = CustomInstrumentsStream(name, session=session)
        return stream



# coding: utf8

from typing import Optional, Iterable, Union

from ....._tools import validate_types
from ._cap_floor_pricing_parameters import PricingParameters
from .._base_definition import BaseDefinition
from ._cap_floor_definition import CapFloorInstrumentDefinition
from ._enums import (
    AdjustInterestToPaymentDate,
    IndexResetType,
    Frequency,
    DateRollingConvention,
    DayCountBasis,
    StubRule,
    BuySell,
    BusinessDayConvention,
)
from ._models import (
    AmortizationItem,
    BarrierDefinitionElement,
)


class Definition(BaseDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
    start_date : str, optional
        The option start date
    end_date : str, optional
        The maturity date of the CapFloor
    tenor : str, optional
        Tenor of the option
    notional_ccy : str, optional
        The ISO code of the notional currency. Mandatory if instrument code or
        instrument style has not been defined. In case an instrument code/style has been
        defined, value may comes from the reference data.
    notional_amount : float, optional
        The notional amount of the leg at the period start date. Optional. By default
        1,000,000 is used.
    index_name : str, optional
        The name of the floating rate index.
    index_tenor : str, optional
        The period code that represents the maturity of the floating rate index.
        Mandatory when the leg is float.
    interest_payment_frequency : Frequency, optional
        The frequency of the interest payments. Optional if an instrument code/style
        have been defined : in that case, value comes from reference data. Otherwise, it
        is mandatory.
    interest_calculation_method : DayCountBasis, str, optional
        The Day Count Basis method used to calculate the coupon interest payments.
        Mandatory.
    payment_business_day_convention : BusinessDayConvention, optional
        The method to adjust dates to a working day. The possible values are:
        - ModifiedFollowing (adjusts dates according to the Modified Following
          convention - next business day unless is it goes into the next month,
          preceeding is used in that  case),
        - NextBusinessDay (adjusts dates according to the Following convention - Next
          Business Day),
        - PreviousBusinessDay (adjusts dates  according to the Preceeding convention -
          Previous Business Day),
        - NoMoving (does not adjust dates),
        - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following
          convention). Optional. In case an instrument code/style has been defined,
          value comes from bond reference data. Otherwise 'ModifiedFollowing' is used.
    payment_roll_convention : DateRollingConvention, optional
        The method to adjust payment dates whn they fall at the end of the month (28th
        of February, 30th, 31st). The possible values are:
        - Last (For setting the calculated date to the last working day),
        - Same (For setting the calculated date to the same day . In this latter case,
          the date may be moved according to the date moving convention if it is a
          non-working day),
        - Last28 (For setting the calculated date to the last working day. 28FEB being
          always considered as the last working day),
        - Same28 (For setting the calculated date to the same day .28FEB being always
          considered as the last working day). Optional. By default 'SameDay' is used.
    index_reset_frequency : Frequency, optional
        The reset frequency in case the leg Type is Float. Optional. By default the
        IndexTenor is used.
    index_reset_type : IndexResetType, optional
        A flag that indicates if the floating rate index is reset before the coupon
        period starts or at the end of the coupon period. The possible values are:
        - InAdvance (resets the index before the start of the interest period),
        - InArrears (resets the index at the end of the interest period). Optional. By
          default 'InAdvance' is used.
    index_fixing_lag : int, optional
        Defines the positive number of working days between the fixing date and the
        start of the coupon period ('InAdvance') or the end of the coupon period
        ('InArrears'). Optional. By default 0 is used.
    amortization_schedule : list of AmortizationItem, optional
        Definition of amortizations
    payment_business_days : str, optional
        A list of coma-separated calendar codes to adjust dates (e.g. 'EMU' or 'USA').
        Optional. By default the calendar associated to NotionalCcy is used.
    adjust_interest_to_payment_date : AdjustInterestToPaymentDate, optional
        A flag that indicates if the coupon dates are adjusted to the payment dates.
        Optional. By default 'false' is used.
    stub_rule : StubRule, optional
        The rule that defines whether coupon roll dates are aligned on the  maturity or
        the issue date. The possible values are:
        - ShortFirstProRata (to create a short period between the start date and the
          first coupon date, and pay a smaller amount of interest for the short
          period.All coupon dates are calculated backward from the maturity date),
        - ShortFirstFull (to create a short period between the start date and the first
          coupon date, and pay a regular coupon on the first coupon date. All coupon
          dates are calculated backward from the maturity date),
        - LongFirstFull (to create a long period between the start date and the second
          coupon date, and pay a regular coupon on the second coupon date. All coupon
          dates are calculated backward from the maturity date),
        - ShortLastProRata (to create a short period between the last payment date and
          maturity, and pay a smaller amount of interest for the short period. All
          coupon dates are calculated forward from the start date). This property may
          also be used in conjunction with firstRegularPaymentDate and
          lastRegularPaymentDate; in that case the following values can be defined:
        - Issue (all dates are aligned on the issue date),
        - Maturity (all dates are aligned on the maturity date). Optional. By default
          'Maturity' is used.
    barrier_definition : BarrierDefinitionElement, optional

    buy_sell : BuySell, optional
        The side of the deal. Possible values:
        - Buy
        - Sell
    annualized_rebate : bool, optional

    cap_digital_payout_percent : float, optional

    cap_strike_percent : float, optional
        Cap leg strike expressed in %
    cms_template : str, optional
        A reference to a common swap contract that represents the underlying swap in
        case of a Constant Maturity Swap contract (CMS). Mandatory for CMS contract.
    floor_digital_payout_percent : float, optional

    floor_strike_percent : float, optional
        Floor leg strike expressed in %
    index_fixing_ric : str, optional
        The RIC that carries the fixing value. This value overrides the RIC associated
        by default with the IndexName and IndexTenor. Optional.
    fields: list of str, optional
        Contains the list of Analytics that the quantitative analytic service will
        compute.
    pricing_parameters : PricingParameters, optional
        The pricing parameters to apply to this instrument. Optional. If pricing
        parameters are not provided at this level parameters defined globally at the
        request level are used. If no pricing parameters are provided globally default
        values apply.
    extended_params : dict, optional
        If necessary other parameters

    Methods
    -------
    get_data(session=session, on_response=on_response)
        Returns a response to the data platform
    get_stream(session=session, api="")
        Get stream object of this definition

    Examples
    --------
    >>> import refinitiv.data.content.ipa.financial_contracts as rdf
    >>> definition = rdf.cap_floor.Definition(
    ...    instrument_tag="CapOnCms",
    ...    stub_rule=rdf.cap_floor.StubRule.MATURITY,
    ...    notional_ccy="USD",
    ...    start_date="2018-06-15",
    ...    end_date="2022-06-15",
    ...    notional_amount=1000000,
    ...    index_name="Composite",
    ...    index_tenor="5Y",
    ...    interest_calculation_method="Dcb_Actual_360",
    ...    interest_payment_frequency=rdf.cap_floor.Frequency.QUARTERLY,
    ...    buy_sell=rdf.cap_floor.BuySell.BUY,
    ...    cap_strike_percent=1,
    ...    pricing_parameters=rdf.cap_floor.PricingParameters(
    ...        skip_first_cap_floorlet=False, valuation_date="2020-02-07"
    ...    ),
    ...    fields=[
    ...        "InstrumentTag",
    ...        "InstrumentDescription",
    ...        "StartDate",
    ...        "EndDate",
    ...        "InterestPaymentFrequency",
    ...        "IndexRic",
    ...        "CapStrikePercent",
    ...        "FloorStrikePercent",
    ...        "NotionalCcy",
    ...        "NotionalAmount",
    ...        "PremiumBp",
    ...        "PremiumPercent",
    ...        "MarketValueInDealCcy",
    ...        "MarketValueInReportCcy",
    ...        "ErrorMessage",
    ...    ],
    ...)
    >>> response = definition.get_data()

     Using get_stream
     >>> response = definition.get_stream()
    """

    def __init__(
        self,
        *,
        instrument_tag: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        notional_ccy: Optional[str] = None,
        notional_amount: Optional[float] = None,
        index_name: Optional[str] = None,
        index_tenor: Optional[str] = None,
        interest_payment_frequency: Union[Frequency, str] = None,
        interest_calculation_method: Union[DayCountBasis, str] = None,
        payment_business_day_convention: Optional[BusinessDayConvention] = None,
        payment_roll_convention: Optional[DateRollingConvention] = None,
        index_reset_frequency: Optional[Frequency] = None,
        index_reset_type: Optional[IndexResetType] = None,
        index_fixing_lag: Optional[int] = None,
        amortization_schedule: Optional[Iterable[AmortizationItem]] = None,
        payment_business_days: Optional[str] = None,
        adjust_interest_to_payment_date: Optional[AdjustInterestToPaymentDate] = None,
        stub_rule: Optional[StubRule] = None,
        barrier_definition: Optional[BarrierDefinitionElement] = None,
        buy_sell: Union[BuySell, str] = None,
        annualized_rebate: Optional[bool] = None,
        cap_digital_payout_percent: Optional[float] = None,
        cap_strike_percent: Optional[float] = None,
        cms_template: Optional[str] = None,
        floor_digital_payout_percent: Optional[float] = None,
        floor_strike_percent: Optional[float] = None,
        index_fixing_ric: Optional[str] = None,
        fields: Optional[Iterable[str]] = None,
        pricing_parameters: Optional[PricingParameters] = None,
        extended_params: Optional[dict] = None,
    ):
        validate_types(index_fixing_lag, [int, type(None)], "index_fixing_lag")

        definition = CapFloorInstrumentDefinition(
            instrument_tag=instrument_tag,
            start_date=start_date,
            end_date=end_date,
            tenor=tenor,
            notional_ccy=notional_ccy,
            notional_amount=notional_amount,
            index_name=index_name,
            index_tenor=index_tenor,
            interest_payment_frequency=interest_payment_frequency,
            interest_calculation_method=interest_calculation_method,
            payment_business_day_convention=payment_business_day_convention,
            payment_roll_convention=payment_roll_convention,
            index_reset_frequency=index_reset_frequency,
            index_reset_type=index_reset_type,
            index_fixing_lag=index_fixing_lag,
            amortization_schedule=amortization_schedule,
            payment_business_days=payment_business_days,
            adjust_interest_to_payment_date=adjust_interest_to_payment_date,
            stub_rule=stub_rule,
            barrier_definition=barrier_definition,
            buy_sell=buy_sell,
            annualized_rebate=annualized_rebate,
            cap_digital_payout_percent=cap_digital_payout_percent,
            cap_strike_percent=cap_strike_percent,
            cms_template=cms_template,
            floor_digital_payout_percent=floor_digital_payout_percent,
            floor_strike_percent=floor_strike_percent,
            index_fixing_ric=index_fixing_ric,
        )
        super().__init__(
            definition=definition,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )



from ..._enums import (
    BenchmarkYieldSelectionMode,
    AdjustInterestToPaymentDate,
    StubRule,
    InterestType,
    Direction,
    IndexCompoundingMethod,
    BusinessDayConvention,
    DateRollingConvention,
    DayCountBasis,
    Frequency,
    PriceSide,
    ProjectedIndexCalculationMethod,
    QuoteFallbackLogic,
    RedemptionDateType,
    Rounding,
    RoundingType,
    YieldType,
    VolatilityType,
    CreditSpreadType,
    DividendType,
    VolatilityTermStructureType,
)



# coding: utf8

__all__ = ["BidAskMid"]

from typing import Optional

from .._object_definition import ObjectDefinition


class BidAskMid(ObjectDefinition):
    """
    Parameters
    ----------
    bid : float, optional

    ask : float, optional

    mid : float, optional

    """

    def __init__(
        self,
        bid: Optional[float] = None,
        ask: Optional[float] = None,
        mid: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.bid = bid
        self.ask = ask
        self.mid = mid

    @property
    def ask(self):
        """
        :return: float
        """
        return self._get_parameter("ask")

    @ask.setter
    def ask(self, value):
        self._set_parameter("ask", value)

    @property
    def bid(self):
        """
        :return: float
        """
        return self._get_parameter("bid")

    @bid.setter
    def bid(self, value):
        self._set_parameter("bid", value)

    @property
    def mid(self):
        """
        :return: float
        """
        return self._get_parameter("mid")

    @mid.setter
    def mid(self, value):
        self._set_parameter("mid", value)



import numpy as np
import pandas as pd


def convert_dict_to_df(data, columns) -> pd.DataFrame:
    if data:
        df = pd.DataFrame(np.array(data), columns=columns)
        if not df.empty:
            df.fillna(pd.NA, inplace=True)
            df = df.convert_dtypes()
    else:
        df = pd.DataFrame([], columns=columns)
    return df


def convert_content_data_to_df(
    content_data: dict, use_field_names_in_headers: bool = True
):
    if "headers" not in content_data or "data" not in content_data:
        return pd.DataFrame()

    if use_field_names_in_headers:
        key = "name"
    else:
        key = "title"

    headers_names = [header[key] for header in content_data["headers"]]
    return convert_dict_to_df(content_data["data"], headers_names)



from typing import TYPE_CHECKING, Any, Union

from ._api_type import APIType
from ._data_provider import DataProvider, default_data_provider
from ._data_type import DataType
from ._endpoint_data_provider import endpoint_data_provider
from ..cfs._cfs_data_provider import (
    cfs_buckets_data_provider,
    cfs_file_sets_data_provider,
    cfs_files_data_provider,
    cfs_packages_data_provider,
    cfs_stream_data_provider,
)
from ..._tools._common import urljoin
from ...content._content_type import ContentType
from ...content.custom_instruments._data_provider import (
    custom_instrument_data_provider,
    custom_instrument_search_data_provider,
    custom_instruments_events_data_provider,
    custom_instruments_interday_summaries_data_provider,
    custom_instruments_intraday_summaries_data_provider,
)
from ...content.esg._esg_data_provider import esg_data_provider
from ...content.estimates._data_provider import estimates_data_provider
from ...content.fundamental_and_reference._data_provider import (
    data_grid_rdp_data_provider,
    data_grid_udf_data_provider,
)
from ...content.historical_pricing._hp_data_provider import (
    hp_events_data_provider,
    hp_summaries_data_provider,
)
from ...content.ipa._curves._curves_data_provider import curves_data_provider
from ...content.ipa._surfaces._surfaces_data_provider import surfaces_data_provider
from ...content.ipa.financial_contracts._contracts_data_provider import (
    contracts_data_provider,
)
from ...content.news._news_data_provider import (
    news_headlines_data_provider_rdp,
    news_headlines_data_provider_udf,
    news_story_data_provider_rdp,
    news_story_data_provider_udf,
)
from ...content.ownership._ownership_data_provider import ownership_data_provider
from ...content.pricing._pricing_content_provider import pricing_data_provider
from ...content.pricing.chain._chains_data_provider import chains_data_provider
from ...content.search._data_provider import (
    search_data_provider,
    lookup_data_provider,
    metadata_data_provider,
)
from ...content.estimates._data_provider import estimates_data_provider
from ...content.pricing._pricing_content_provider import pricing_data_provider
from ...content.filings._data_provider import filings_data_provider
from ..._tools._common import urljoin

if TYPE_CHECKING:
    from ..._configure import _RDPConfig

data_provider_by_data_type = {
    ContentType.CHAINS: chains_data_provider,
    ContentType.CONTRACTS: contracts_data_provider,
    ContentType.CUSTOM_INSTRUMENTS_EVENTS: custom_instruments_events_data_provider,
    ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS: custom_instrument_data_provider,
    ContentType.CUSTOM_INSTRUMENTS_INTERDAY_SUMMARIES: custom_instruments_interday_summaries_data_provider,
    ContentType.CUSTOM_INSTRUMENTS_INTRADAY_SUMMARIES: custom_instruments_intraday_summaries_data_provider,
    ContentType.CUSTOM_INSTRUMENTS_SEARCH: custom_instrument_search_data_provider,
    ContentType.DATA_GRID_RDP: data_grid_rdp_data_provider,
    ContentType.DATA_GRID_UDF: data_grid_udf_data_provider,
    ContentType.DEFAULT: default_data_provider,
    ContentType.DISCOVERY_LOOKUP: lookup_data_provider,
    ContentType.DISCOVERY_METADATA: metadata_data_provider,
    ContentType.DISCOVERY_SEARCH: search_data_provider,
    ContentType.ESG_BASIC_OVERVIEW: esg_data_provider,
    ContentType.ESG_FULL_MEASURES: esg_data_provider,
    ContentType.ESG_FULL_SCORES: esg_data_provider,
    ContentType.ESG_STANDARD_MEASURES: esg_data_provider,
    ContentType.ESG_STANDARD_SCORES: esg_data_provider,
    ContentType.ESG_UNIVERSE: esg_data_provider,
    ContentType.ESTIMATES_VIEW_ACTUALS_ANNUAL: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_ACTUALS_INTERIM: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_ACTUALS_KPI_ANNUAL: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_ACTUALS_KPI_INTERIM: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_ANNUAL: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_NON_PERIODIC_MEASURES: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_ANNUAL: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_INTERIM: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_RECOMMENDATIONS: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_INTERIM: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_ANNUAL: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_HISTORICAL_SNAPSHOTS_KPI: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_INTERIM: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_NON_PERIODIC_MEASURES: estimates_data_provider,
    ContentType.ESTIMATES_VIEW_SUMMARY_RECOMMENDATIONS: estimates_data_provider,
    ContentType.FILINGS_RETRIEVAL: filings_data_provider,
    ContentType.FORWARD_CURVE: curves_data_provider,
    ContentType.HISTORICAL_PRICING_EVENTS: hp_events_data_provider,
    ContentType.HISTORICAL_PRICING_INTERDAY_SUMMARIES: hp_summaries_data_provider,
    ContentType.HISTORICAL_PRICING_INTRADAY_SUMMARIES: hp_summaries_data_provider,
    ContentType.NEWS_HEADLINES_RDP: news_headlines_data_provider_rdp,
    ContentType.NEWS_HEADLINES_UDF: news_headlines_data_provider_udf,
    ContentType.NEWS_STORY_RDP: news_story_data_provider_rdp,
    ContentType.NEWS_STORY_UDF: news_story_data_provider_udf,
    ContentType.OWNERSHIP_CONSOLIDATED_BREAKDOWN: ownership_data_provider,
    ContentType.OWNERSHIP_CONSOLIDATED_CONCENTRATION: ownership_data_provider,
    ContentType.OWNERSHIP_CONSOLIDATED_INVESTORS: ownership_data_provider,
    ContentType.OWNERSHIP_CONSOLIDATED_RECENT_ACTIVITY: ownership_data_provider,
    ContentType.OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_HISTORY_REPORT: ownership_data_provider,
    ContentType.OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_REPORT: ownership_data_provider,
    ContentType.OWNERSHIP_CONSOLIDATED_TOP_N_CONCENTRATION: ownership_data_provider,
    ContentType.OWNERSHIP_FUND_BREAKDOWN: ownership_data_provider,
    ContentType.OWNERSHIP_FUND_CONCENTRATION: ownership_data_provider,
    ContentType.OWNERSHIP_FUND_HOLDINGS: ownership_data_provider,
    ContentType.OWNERSHIP_FUND_INVESTORS: ownership_data_provider,
    ContentType.OWNERSHIP_FUND_RECENT_ACTIVITY: ownership_data_provider,
    ContentType.OWNERSHIP_FUND_SHAREHOLDERS_HISTORY_REPORT: ownership_data_provider,
    ContentType.OWNERSHIP_FUND_SHAREHOLDERS_REPORT: ownership_data_provider,
    ContentType.OWNERSHIP_FUND_TOP_N_CONCENTRATION: ownership_data_provider,
    ContentType.OWNERSHIP_INSIDER_SHAREHOLDERS_REPORT: ownership_data_provider,
    ContentType.OWNERSHIP_INSIDER_TRANSACTION_REPORT: ownership_data_provider,
    ContentType.OWNERSHIP_INVESTOR_HOLDINGS: ownership_data_provider,
    ContentType.OWNERSHIP_ORG_INFO: ownership_data_provider,
    ContentType.PRICING: pricing_data_provider,
    ContentType.SURFACES: surfaces_data_provider,
    ContentType.ZC_CURVE_DEFINITIONS: curves_data_provider,
    ContentType.ZC_CURVES: curves_data_provider,
    DataType.CFS_BUCKETS: cfs_buckets_data_provider,
    DataType.CFS_FILE_SETS: cfs_file_sets_data_provider,
    DataType.CFS_FILES: cfs_files_data_provider,
    DataType.CFS_PACKAGES: cfs_packages_data_provider,
    DataType.CFS_STREAM: cfs_stream_data_provider,
    DataType.ENDPOINT: endpoint_data_provider,
}

api_config_key_by_api_type = {
    APIType.CFS: "apis.file-store",
    APIType.CURVES_AND_SURFACES: "apis.data.quantitative-analytics-curves-and-surfaces",
    APIType.FINANCIAL_CONTRACTS: "apis.data.quantitative-analytics-financial-contracts",
    APIType.HISTORICAL_PRICING: "apis.data.historical-pricing",
    APIType.ESG: "apis.data.environmental-social-governance",
    APIType.PRICING: "apis.data.pricing",
    APIType.OWNERSHIP: "apis.data.ownership",
    APIType.CHAINS: "apis.data.pricing",
    APIType.DATA_GRID: "apis.data.datagrid",
    APIType.NEWS: "apis.data.news",
    APIType.DISCOVERY: "apis.discovery.search",
    APIType.ESTIMATES: "apis.data.estimates",
    APIType.CUSTOM_INSTRUMENTS: "apis.data.custom-instruments",
    APIType.FILINGS: "apis.data.filings",
}

api_type_by_data_type = {
    DataType.CFS_BUCKETS: APIType.CFS,
    DataType.CFS_FILE_SETS: APIType.CFS,
    DataType.CFS_FILES: APIType.CFS,
    DataType.CFS_PACKAGES: APIType.CFS,
    DataType.CFS_STREAM: APIType.CFS,
    ContentType.FORWARD_CURVE: APIType.CURVES_AND_SURFACES,
    ContentType.ZC_CURVES: APIType.CURVES_AND_SURFACES,
    ContentType.ZC_CURVE_DEFINITIONS: APIType.CURVES_AND_SURFACES,
    ContentType.SURFACES: APIType.CURVES_AND_SURFACES,
    ContentType.CONTRACTS: APIType.FINANCIAL_CONTRACTS,
    ContentType.HISTORICAL_PRICING_EVENTS: APIType.HISTORICAL_PRICING,
    ContentType.HISTORICAL_PRICING_INTERDAY_SUMMARIES: APIType.HISTORICAL_PRICING,
    ContentType.HISTORICAL_PRICING_INTRADAY_SUMMARIES: APIType.HISTORICAL_PRICING,
    ContentType.ESG_STANDARD_SCORES: APIType.ESG,
    ContentType.ESG_STANDARD_MEASURES: APIType.ESG,
    ContentType.ESG_FULL_MEASURES: APIType.ESG,
    ContentType.ESG_FULL_SCORES: APIType.ESG,
    ContentType.ESG_BASIC_OVERVIEW: APIType.ESG,
    ContentType.ESG_UNIVERSE: APIType.ESG,
    ContentType.PRICING: APIType.PRICING,
    ContentType.OWNERSHIP_CONSOLIDATED_BREAKDOWN: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_CONSOLIDATED_CONCENTRATION: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_CONSOLIDATED_INVESTORS: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_CONSOLIDATED_RECENT_ACTIVITY: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_HISTORY_REPORT: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_REPORT: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_CONSOLIDATED_TOP_N_CONCENTRATION: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_FUND_CONCENTRATION: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_FUND_BREAKDOWN: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_FUND_INVESTORS: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_FUND_RECENT_ACTIVITY: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_FUND_SHAREHOLDERS_HISTORY_REPORT: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_FUND_SHAREHOLDERS_REPORT: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_FUND_TOP_N_CONCENTRATION: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_FUND_HOLDINGS: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_INSIDER_SHAREHOLDERS_REPORT: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_INSIDER_TRANSACTION_REPORT: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_INVESTOR_HOLDINGS: APIType.OWNERSHIP,
    ContentType.OWNERSHIP_ORG_INFO: APIType.OWNERSHIP,
    ContentType.CHAINS: APIType.CHAINS,
    ContentType.DATA_GRID_RDP: APIType.DATA_GRID,
    ContentType.DATA_GRID_UDF: APIType.DATA_GRID,
    ContentType.NEWS_HEADLINES_RDP: APIType.NEWS,
    ContentType.NEWS_HEADLINES_UDF: APIType.NEWS,
    ContentType.NEWS_STORY_RDP: APIType.NEWS,
    ContentType.NEWS_STORY_UDF: APIType.NEWS,
    ContentType.DISCOVERY_SEARCH: APIType.DISCOVERY,
    ContentType.DISCOVERY_LOOKUP: APIType.DISCOVERY,
    ContentType.DISCOVERY_METADATA: APIType.DISCOVERY,
    ContentType.ESTIMATES_VIEW_ACTUALS_ANNUAL: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_ACTUALS_INTERIM: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_ACTUALS_KPI_ANNUAL: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_ACTUALS_KPI_INTERIM: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_ANNUAL: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_NON_PERIODIC_MEASURES: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_ANNUAL: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_INTERIM: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_RECOMMENDATIONS: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_INTERIM: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_NON_PERIODIC_MEASURES: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_RECOMMENDATIONS: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_ANNUAL: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_HISTORICAL_SNAPSHOTS_KPI: APIType.ESTIMATES,
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_INTERIM: APIType.ESTIMATES,
    ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS: APIType.CUSTOM_INSTRUMENTS,
    ContentType.CUSTOM_INSTRUMENTS_SEARCH: APIType.CUSTOM_INSTRUMENTS,
    ContentType.CUSTOM_INSTRUMENTS_EVENTS: APIType.CUSTOM_INSTRUMENTS,
    ContentType.CUSTOM_INSTRUMENTS_INTRADAY_SUMMARIES: APIType.CUSTOM_INSTRUMENTS,
    ContentType.CUSTOM_INSTRUMENTS_INTERDAY_SUMMARIES: APIType.CUSTOM_INSTRUMENTS,
    ContentType.FILINGS_RETRIEVAL: APIType.FILINGS,
}

url_config_key_by_data_type = {
    DataType.CFS_BUCKETS: "endpoints.buckets",
    DataType.CFS_FILE_SETS: "endpoints.file-sets",
    DataType.CFS_FILES: "endpoints.files",
    DataType.CFS_PACKAGES: "endpoints.packages",
    DataType.CFS_STREAM: "endpoints.files",
    ContentType.FORWARD_CURVE: "endpoints.forward-curves",
    ContentType.ZC_CURVES: "endpoints.zc-curves",
    ContentType.ZC_CURVE_DEFINITIONS: "endpoints.zc-curve-definitions",
    ContentType.SURFACES: "endpoints.surfaces",
    ContentType.CONTRACTS: "endpoints.financial-contracts",
    ContentType.HISTORICAL_PRICING_EVENTS: "endpoints.events",
    ContentType.HISTORICAL_PRICING_INTERDAY_SUMMARIES: "endpoints.interday-summaries",
    ContentType.HISTORICAL_PRICING_INTRADAY_SUMMARIES: "endpoints.intraday-summaries",
    ContentType.ESG_STANDARD_SCORES: "endpoints.scores-standard",
    ContentType.ESG_STANDARD_MEASURES: "endpoints.measures-standard",
    ContentType.ESG_FULL_MEASURES: "endpoints.measures-full",
    ContentType.ESG_FULL_SCORES: "endpoints.scores-full",
    ContentType.ESG_BASIC_OVERVIEW: "endpoints.basic",
    ContentType.ESG_UNIVERSE: "endpoints.universe",
    ContentType.PRICING: "endpoints.snapshots",
    ContentType.OWNERSHIP_CONSOLIDATED_BREAKDOWN: "endpoints.consolidated.breakdown",
    ContentType.OWNERSHIP_CONSOLIDATED_CONCENTRATION: "endpoints.consolidated.concentration",
    ContentType.OWNERSHIP_CONSOLIDATED_INVESTORS: "endpoints.consolidated.investors",
    ContentType.OWNERSHIP_CONSOLIDATED_RECENT_ACTIVITY: "endpoints.consolidated.recent-activity",
    ContentType.OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_HISTORY_REPORT: "endpoints.consolidated.shareholders-history-report",
    ContentType.OWNERSHIP_CONSOLIDATED_SHAREHOLDERS_REPORT: "endpoints.consolidated.shareholders-report",
    ContentType.OWNERSHIP_CONSOLIDATED_TOP_N_CONCENTRATION: "endpoints.consolidated.top-n-concentration",
    ContentType.OWNERSHIP_FUND_CONCENTRATION: "endpoints.fund.concentration",
    ContentType.OWNERSHIP_FUND_BREAKDOWN: "endpoints.fund.breakdown",
    ContentType.OWNERSHIP_FUND_INVESTORS: "endpoints.fund.investors",
    ContentType.OWNERSHIP_FUND_RECENT_ACTIVITY: "endpoints.fund.recent-activity",
    ContentType.OWNERSHIP_FUND_SHAREHOLDERS_HISTORY_REPORT: "endpoints.fund.shareholders-history-report",
    ContentType.OWNERSHIP_FUND_SHAREHOLDERS_REPORT: "endpoints.fund.shareholders-report",
    ContentType.OWNERSHIP_FUND_TOP_N_CONCENTRATION: "endpoints.fund.top-n-concentration",
    ContentType.OWNERSHIP_FUND_HOLDINGS: "endpoints.fund.holdings",
    ContentType.OWNERSHIP_INSIDER_SHAREHOLDERS_REPORT: "endpoints.insider.shareholders-report",
    ContentType.OWNERSHIP_INSIDER_TRANSACTION_REPORT: "endpoints.insider.transaction-report",
    ContentType.OWNERSHIP_INVESTOR_HOLDINGS: "endpoints.investor.holdings",
    ContentType.OWNERSHIP_ORG_INFO: "endpoints.org-info",
    ContentType.CHAINS: "endpoints.chains",
    ContentType.DATA_GRID_RDP: "endpoints.standard",
    ContentType.DATA_GRID_UDF: "endpoints.standard",
    ContentType.NEWS_HEADLINES_RDP: "endpoints.headlines",
    ContentType.NEWS_HEADLINES_UDF: "endpoints.headlines",
    ContentType.NEWS_STORY_RDP: "endpoints.stories",
    ContentType.NEWS_STORY_UDF: "endpoints.stories",
    ContentType.DISCOVERY_SEARCH: "endpoints.search",
    ContentType.DISCOVERY_LOOKUP: "endpoints.lookup",
    ContentType.DISCOVERY_METADATA: "endpoints.metadata",
    ContentType.ESTIMATES_VIEW_ACTUALS_ANNUAL: "endpoints.view-actuals.annual",
    ContentType.ESTIMATES_VIEW_ACTUALS_INTERIM: "endpoints.view-actuals.interim",
    ContentType.ESTIMATES_VIEW_ACTUALS_KPI_ANNUAL: "endpoints.view-actuals-kpi.annual",
    ContentType.ESTIMATES_VIEW_ACTUALS_KPI_INTERIM: "endpoints.view-actuals-kpi.interim",
    ContentType.ESTIMATES_VIEW_SUMMARY_ANNUAL: "endpoints.view-summary.annual",
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_NON_PERIODIC_MEASURES: "endpoints.view-summary.historical-snapshots-non-periodic-measures",
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_ANNUAL: "endpoints.view-summary.historical-snapshots-periodic-measures-annual",
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_PERIODIC_MEASURES_INTERIM: "endpoints.view-summary.historical-snapshots-periodic-measures-interim",
    ContentType.ESTIMATES_VIEW_SUMMARY_HISTORICAL_SNAPSHOTS_RECOMMENDATIONS: "endpoints.view-summary.historical-snapshots-recommendations",
    ContentType.ESTIMATES_VIEW_SUMMARY_INTERIM: "endpoints.view-summary.interim",
    ContentType.ESTIMATES_VIEW_SUMMARY_NON_PERIODIC_MEASURES: "endpoints.view-summary.non-periodic-measures",
    ContentType.ESTIMATES_VIEW_SUMMARY_RECOMMENDATIONS: "endpoints.view-summary.recommendations",
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_ANNUAL: "endpoints.view-summary-kpi.annual",
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_HISTORICAL_SNAPSHOTS_KPI: "endpoints.view-summary-kpi.historical-snapshots-kpi",
    ContentType.ESTIMATES_VIEW_SUMMARY_KPI_INTERIM: "endpoints.view-summary-kpi.interim",
    ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS: "endpoints.instruments",
    ContentType.CUSTOM_INSTRUMENTS_SEARCH: "endpoints.search",
    ContentType.CUSTOM_INSTRUMENTS_EVENTS: "endpoints.events",
    ContentType.CUSTOM_INSTRUMENTS_INTERDAY_SUMMARIES: "endpoints.interday-summaries",
    ContentType.CUSTOM_INSTRUMENTS_INTRADAY_SUMMARIES: "endpoints.intraday-summaries",
    ContentType.FILINGS_RETRIEVAL: "endpoints.retrieval",
}


def _get_api_config_key(data_type: DataType) -> str:
    api_type = api_type_by_data_type.get(data_type)
    api_config_key = api_config_key_by_api_type.get(api_type)
    return api_config_key


def _get_url_config_key(data_type: DataType) -> str:
    return url_config_key_by_data_type.get(data_type)


def get_api_config(
    data_type: Union[DataType, ContentType], config: "_RDPConfig"
) -> Union[dict, Any]:
    api_config_key = _get_api_config_key(data_type)
    api_config = config.get(api_config_key)

    if api_config is None:
        raise AttributeError(
            f"Cannot find api_key, data_type={data_type} by key={api_config_key}"
        )

    return api_config


def get_base_url(data_type: DataType, config: "_RDPConfig") -> str:
    """
    Parameters
    ----------
    config: _RDPConfig
    data_type: DataType

    Returns
    -------
    string
    """
    api_config = get_api_config(data_type, config)
    base_url = api_config.get("url")
    return base_url


def get_url(
    data_type: Union[DataType, ContentType],
    config: "_RDPConfig",
    request_mode: str = None,
) -> str:
    """
    Parameters
    ----------
    config: _RDPConfig
    data_type: DataType
    request_mode: str in ["sync", "async"]

    Returns
    -------
    string
    """
    api_config = get_api_config(data_type, config)
    url_config_key = _get_url_config_key(data_type)

    if request_mode is None:
        request_mode = "sync"

    if request_mode not in ["sync", "async"]:
        raise AttributeError(f"Request mode not in ['sync', 'async'].")

    # Test if url_content is a json with sync/async endpoints
    url_config_key_with_request_mode = ".".join([url_config_key, request_mode])
    content_url = api_config.get(url_config_key_with_request_mode)

    if content_url is None:
        # then test if content_url is a single endpoint
        content_url = api_config.get(url_config_key)

    if content_url is None:
        raise AttributeError(
            f"Cannot find content_url, data_type={data_type} by key={url_config_key}"
        )

    base_url = api_config.get("url")
    url = urljoin(base_url, content_url)

    return url


def make_provider(data_type: Union[DataType, ContentType], **_) -> DataProvider:
    """
    Parameters
    ----------
    data_type: DataType

    Returns
    -------
    DataProvider
    """
    data_provider = data_provider_by_data_type.get(data_type)

    if data_provider is None:
        raise AttributeError(f"Cannot get data provider by content type: {data_type}")

    return data_provider



# coding: utf8
# contract_gen 2020-06-16 10:26:10.695191

__all__ = ["VolatilityAdjustmentType"]

from enum import Enum, unique

from ._common_tools import _convert_to_str, _normalize


@unique
class VolatilityAdjustmentType(Enum):
    CONSTANT_CAP = "ConstantCap"
    CONSTANT_CAPLET = "ConstantCaplet"
    NORMALIZED_CAP = "NormalizedCap"
    NORMALIZED_CAPLET = "NormalizedCaplet"
    PB_UNDEFINED = "PbUndefined"
    SHIFTED_CAP = "ShiftedCap"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            VolatilityAdjustmentType, _VOLATILITY_ADJUSTMENT_TYPE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _VOLATILITY_ADJUSTMENT_TYPE_VALUES_IN_LOWER_BY_VOLATILITY_ADJUSTMENT_TYPE,
            some,
        )


_VOLATILITY_ADJUSTMENT_TYPE_VALUES = tuple(t.value for t in VolatilityAdjustmentType)
_VOLATILITY_ADJUSTMENT_TYPE_VALUES_IN_LOWER_BY_VOLATILITY_ADJUSTMENT_TYPE = {
    name.lower(): item for name, item in VolatilityAdjustmentType.__members__.items()
}



__all__ = (
    "Definition",
    "Session",
    "set_default",
    "get_default",
    "EventCode",
    "desktop",
    "platform",
)

from .._core.session._session_definition import Definition
from .._core.session._session import Session
from .._core.session.event_code import EventCode
from .._core.session._default_session_manager import set_default, get_default
from . import desktop, platform



# coding: utf8

__all__ = ["ExtrapolationMode"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class ExtrapolationMode(Enum):
    CONSTANT = "Constant"
    LINEAR = "Linear"
    NONE = "None"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(ExtrapolationMode, _EXTRAPOLATION_MODE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _EXTRAPOLATION_MODE_VALUES_IN_LOWER_BY_EXTRAPOLATION_MODE, some
        )


_EXTRAPOLATION_MODE_VALUES = tuple(t.value for t in ExtrapolationMode)
_EXTRAPOLATION_MODE_VALUES_IN_LOWER_BY_EXTRAPOLATION_MODE = {
    name.lower(): item for name, item in ExtrapolationMode.__members__.items()
}



from ._base_definition import BaseDefinition
from .._data._data_type import DataType


class CFSStream(BaseDefinition):
    def __init__(self, id):
        super().__init__(data_type=DataType.CFS_STREAM, id=id)



# coding: utf8

from typing import Optional, Iterable

from ._bermudan_swaption_definition import BermudanSwaptionDefinition
from ._swaption_definition import SwaptionInstrumentDefinition
from ._swaption_pricing_parameters import PricingParameters
from .. import swap
from .._base_definition import BaseDefinition
from ..._enums import BuySell, CallPut, ExerciseStyle, SwaptionSettlementType
from ....._tools import create_repr


class Definition(BaseDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument. It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported.
    end_date : str, optional
        Expiry date of the option.
    tenor : str, optional
        Tenor of the option.
    bermudan_swaption_definition : BermudanSwaptionDefinition, optional

    buy_sell : BuySell
        The direction of the trade in terms of the deal currency.
    call_put : CallPut
        Tells if the option is a call or a put.
    exercise_style : ExerciseStyle
        EURO or BERM or AMER.
    settlement_type : SwaptionSettlementType, optional
        The settlement type of the option if the option is exercised
    underlying_definition : swap.Definition
        The details of the underlying swap.
    strike_percent : float, optional
        Strike percent of the option expressed in % format.
    fields: list of str, optional
        Contains the list of Analytics that the quantitative analytic service will
        compute.
    pricing_parameters : PricingParameters, optional
        The pricing parameters to apply to this instrument. If pricing parameters
        are not provided at this level parameters defined globally at the request
        level are used. If no pricing parameters are provided globally default
        values apply.
    extended_params : dict, optional
        If necessary other parameters.

    Methods
    -------
    get_data(session=session, on_response=on_response, async_mode=None)
        Returns a response to the data platform
    get_data_async(session=session, on_response=on_response, async_mode=None)
        Returns a response to the async data platform
    get_stream(session=session, api="")
        Get stream quantitative analytic service subscription

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> definition = rdf.swaption.Definition(
     ...   buy_sell=rdf.swaption.BuySell.BUY,
     ...   call_put=rdf.swaption.CallPut.CALL,
     ...   exercise_style=rdf.swaption.ExerciseStyle.BERM,
     ...   underlying_definition=rdf.swap.Definition(tenor="3Y", template="EUR_AB6E"),
     ...)
     >>> response = definition.get_data()

    Using get_stream
     >>> stream = definition.get_stream()
     >>> stream.open()
    """

    def __init__(
        self,
        *,
        instrument_tag: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        bermudan_swaption_definition: Optional[BermudanSwaptionDefinition] = None,
        buy_sell: BuySell = None,
        call_put: CallPut = None,
        exercise_style: ExerciseStyle = None,
        settlement_type: Optional[SwaptionSettlementType] = None,
        underlying_definition: swap.Definition = None,
        strike_percent: Optional[float] = None,
        fields: Optional[Iterable[str]] = None,
        pricing_parameters: Optional[PricingParameters] = None,
        extended_params: Optional[dict] = None,
    ):
        self.underlying_definition = underlying_definition
        definition = SwaptionInstrumentDefinition(
            instrument_tag=instrument_tag,
            end_date=end_date,
            tenor=tenor,
            bermudan_swaption_definition=bermudan_swaption_definition,
            buy_sell=buy_sell,
            call_put=call_put,
            exercise_style=exercise_style,
            settlement_type=settlement_type,
            underlying_definition=underlying_definition,
            strike_percent=strike_percent,
        )
        super().__init__(
            definition=definition,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="content.ipa.financial_contracts.swaption",
            content=f"{{underlying_definition='{self.underlying_definition}'}}",
        )



__all__ = (
    "AssetClass",
    "CalendarAdjustment",
    "CompoundingType",
    "DayCountBasis",
    "Definition",
    "ExtrapolationMode",
    "MarketDataAccessDeniedFallback",
    "PriceSide",
    "RiskType",
    "ZcCurveDefinitions",
    "ZcCurveParameters",
    "ZcInterpolationMode",
)

from ._definition import Definition
from ._enums import (
    DayCountBasis,
    CalendarAdjustment,
    ZcInterpolationMode,
    PriceSide,
    MarketDataAccessDeniedFallback,
    CompoundingType,
    ExtrapolationMode,
    RiskType,
    AssetClass,
)
from ..._curves import ZcCurveDefinitions, ZcCurveParameters



# coding: utf8

from typing import Optional, Iterable

from ._enums import FxCrossType
from ._fx_cross_definition import FxCrossInstrumentDefinition
from ._fx_cross_leg_definition import LegDefinition
from ._fx_cross_pricing_parameters import PricingParameters
from .._base_definition import BaseDefinition


class Definition(BaseDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
    legs : list of LegDefinition, optional
        Extra parameters to describe further the contract. 1 leg is mandatory for
        Forwards and NDFs contracts. 2 legs are required for Swaps, and FwdFwdSwaps
        contracts. Optional for Spot contracts.
    fx_cross_type : FxCrossType, optional
        The type of the Fx Cross instrument :  'FxSpot', 'FxForward',
        'FxNonDeliverableForward', 'FxSwap', 'MultiLeg' or 'FxForwardForward'.
        Mandatory.
    fx_cross_code : str, optional
        The ISO code of the cross currency (e.g. 'EURCHF'). Mandatory.
    ndf_fixing_settlement_ccy : str, optional
        In case of a NDF contract, the ISO code of the settlement currency (e.g. 'EUR'
        ). Optional.
    reference_spot_rate : float, optional
        Contractual Spot Rate the counterparties agreed. It is used to compute the
        traded_cross_rate as 'reference_spot_rate + traded_swap_points /
        FxSwapPointScalingFactor'. In the case of a "FxSwap" contract, it is also used
        to compute  nearLeg.ContraAmount from nearLeg.DealAmount as
        'nearLeg.ContraAmount = nearLeg.DealAmount *  (reference_spot_rate /
        FxCrossScalingFactor)'. Optional. Default value is null. In that case
        traded_cross_rate and Leg ContraAmount may not be computed.
    traded_cross_rate : float, optional
        The contractual exchange rate agreed by the two counterparties.  It is used to
        compute the ContraAmount if the amount is not filled.  In the case of a
        'FxForward' and 'FxNonDeliverableForward' contract : ContraAmount is computed as
        'DealAmount x traded_cross_rate / FxCrossScalingFactor'. In the case of a
        'FxSwap' contract : farLeg.ContraAmount is computed as 'nearLeg.DealAmount x
        traded_cross_rate / FxCrossScalingFactor'. Optional. Default value is null. It
        emans that if both ContraAmount and traded_cross_rate are sot set, market value
        cannot be computed.
    traded_swap_points : float, optional
        Contractual forward points agreed by the two counterparties. It is used to
        compute the traded_cross_rate as 'reference_spot_rate + traded_swap_points /
        FxSwapPointScalingFactor'. Optional. Default value is null. In that case
        traded_cross_rate and Leg ContraAmount may not be computed.
    fields: list of str, optional
        Contains the list of Analytics that the quantitative analytic service will
        compute.
    pricing_parameters : PricingParameters, optional
        The pricing parameters to apply to this instrument. Optional. If pricing
        parameters are not provided at this level parameters defined globally at the
        request level are used. If no pricing parameters are provided globally default
        values apply.
    extended_params : dict, optional
        If necessary other parameters

    Methods
    -------
    get_data(session=session, on_response=on_response)
        Returns a response to the data platform
    get_stream(session=session, api="")
        Get stream quantitative analytic service subscription

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> definition = rdf.cross.Definition(
     ...     instrument_tag="00102700008910C",
     ...     fx_cross_type=rdf.cross.FxCrossType.FX_FORWARD,
     ...     fx_cross_code="USDEUR",
     ...     legs=[rdf.cross.LegDefinition(end_date="2015-04-09T00:00:00Z")],
     ...     pricing_parameters=rdf.cross.PricingParameters(
     ...         valuation_date="2015-02-02T00:00:00Z",
     ...         price_side=rdf.cross.PriceSide.MID,
     ...     ),
     ...     fields=[
     ...         "InstrumentTag",
     ...         "ValuationDate",
     ...         "InstrumentDescription",
     ...         "FxOutrightCcy1Ccy2",
     ...     ],
     ... )
     >>> response = definition.get_data()
     >>> response.data.df

     Using get_stream
     >>> response = definition.get_stream()
    """

    def __init__(
        self,
        *,
        instrument_tag: Optional[str] = None,
        legs: Optional[Iterable[LegDefinition]] = None,
        fx_cross_type: Optional[FxCrossType] = None,
        fx_cross_code: Optional[str] = None,
        ndf_fixing_settlement_ccy: Optional[str] = None,
        reference_spot_rate: Optional[float] = None,
        traded_cross_rate: Optional[float] = None,
        traded_swap_points: Optional[float] = None,
        fields: Optional[Iterable[str]] = None,
        pricing_parameters: Optional[PricingParameters] = None,
        extended_params: Optional[dict] = None,
    ):
        definition = FxCrossInstrumentDefinition(
            fx_cross_type=fx_cross_type,
            legs=legs,
            fx_cross_code=fx_cross_code,
            instrument_tag=instrument_tag,
            ndf_fixing_settlement_ccy=ndf_fixing_settlement_ccy,
            reference_spot_rate=reference_spot_rate,
            traded_cross_rate=traded_cross_rate,
            traded_swap_points=traded_swap_points,
        )
        super().__init__(
            definition=definition,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )



# coding: utf8

from ._enums import ExerciseScheduleType
from ..._object_definition import ObjectDefinition


class BermudanSwaptionDefinition(ObjectDefinition):
    def __init__(
        self,
        exercise_schedule=None,
        exercise_schedule_type=None,
        notification_days=None,
    ):
        super().__init__()
        self.exercise_schedule = exercise_schedule
        self.exercise_schedule_type = exercise_schedule_type
        self.notification_days = notification_days

    @property
    def exercise_schedule(self):
        """
        Overridable exercise schedule
        :return: list string
        """
        return self._get_list_parameter(str, "exerciseSchedule")

    @exercise_schedule.setter
    def exercise_schedule(self, value):
        self._set_list_parameter(str, "exerciseSchedule", value)

    @property
    def exercise_schedule_type(self):
        """
        :return: enum ExerciseScheduleType
        """
        return self._get_enum_parameter(ExerciseScheduleType, "exerciseScheduleType")

    @exercise_schedule_type.setter
    def exercise_schedule_type(self, value):
        self._set_enum_parameter(ExerciseScheduleType, "exerciseScheduleType", value)

    @property
    def notification_days(self):
        """
        :return: int
        """
        return self._get_parameter("notificationDays")

    @notification_days.setter
    def notification_days(self, value):
        self._set_parameter("notificationDays", value)



# coding: utf-8

import json

from . import _json_requests as json_requests
from ._tools import is_string_type, get_default_session, convert_content_data_to_df_udf
from .._tools._datetime import to_datetime

News_Headlines_UDF_endpoint = "News_Headlines"
News_Story_UDF_endpoint = "News_Story"


def get_news_headlines(
    query=None,
    count=10,
    repository="NewsWire",
    date_from=None,
    date_to=None,
    raw_output=False,
    debug=False,
):
    """
    Returns a list of news headlines

    Parameters
    ----------
    query: string, optional
        News headlines search criteria.
        The text can contain RIC codes, company names, country names and
        operators (AND, OR, NOT, IN, parentheses and quotes for explicit search…).

        Tip: Append 'R:' in front of RIC names to improve performance.

        Default: Top News written in English

    count: int, optional
        Max number of headlines retrieved.
        Value Range: [1-100].
        Default: 10

    repository: string, list of strings, optional
        Possible values:
            - "NewsWire"
            - "NewsRoom"
            - "WebNews"
        For "NewsRoom" and "WebNews" repositories a query must be defined.

    date_from: string or datetime, optional
        Beginning of date range.
        String format is: '%Y-%m-%dT%H:%M:%S'. e.g. '2016-01-20T15:04:05'.

    date_to: string or datetime, optional
        End of date range.
        String format is: '%Y-%m-%dT%H:%M:%S'. e.g. '2016-01-20T15:04:05'.

    raw_output: boolean, optional
        Set this parameter to True to get the data in json format
        if set to False, the legacy will return a data frame
        Default: False

    debug: boolean, optional
        When set to True, the json request and response are printed.
        Default: False

    Returns
    -------
    pandas.DataFrame
        Returns a DataFrame of news headlines with the following columns:

        - Index               : Timestamp of the publication time
        - version_created     : Date of the latest update on the news
        - text                : Text of the Headline
        - story_id            : Identifier to be used to retrieve the full story using the get_news_story legacy
        - source_code         : Second news identifier

    Raises
    ----------
    Exception
        If http request fails or if server returns an error
    AttributeError
        If a parameter type is wrong

    Examples
    --------
    >> import refinitiv.data as ek
    >> ek.set_app_key('set your app id here')
    >> headlines = ek.get_news_headlines("R:MSFT.O", 2)
    >> headlines
                                    versionCreated                                              text \
    2016-04-13 18:28:57.000 2.016-04-13 18:28:59.001 RBC Applies Blockchain as a Loyalty Boost<MSFT...
    2016-04-13 17:28:21.577 2016-04-13 17:28:21.671 UPDATE 2-Long-stalled email privacy bill advan...
                                                                    storyId
    2016-04-13 18:28:57.000    urn:newsml:reuters.com:20160413:nNRA1uxh03:1
    2016-04-13 17:28:21.577    urn:newsml:reuters.com:20160413:nL2N17G16Q:2

    >> headlines = ek.get_news_headlines("R:MSFT.O IN FRANCE")
    >> headlines = ek.get_news_headlines("R:MSFT.O IN FRANCE IN ENGLISH", count=5)
    >> headlines = ek.get_news_headlines("OBA* OR CLINTON IN ENGLISH", count=5)
    """
    logger = get_default_session().logger()

    if query is None:
        query = "Topic:TOPALL and Language:LEN"

    # check parameters type and values
    if not is_string_type(query):
        error_msg = "query must be a string"
        logger.error(error_msg)
        raise ValueError(error_msg)

    # query string must be formated as a "" string containing '' substrings
    #   (and not a '' string containing "" substrings)
    query = query.replace('"', "'")

    # validate query JSON format
    test_query = '{"query":"' + query + '"}'
    try:
        json.loads(str(test_query))
    except ValueError as error:
        error_msg = "query {} has invalid format. {}".format(test_query, str(error))
        logger.debug(error_msg)
        raise ValueError(error_msg)

    if type(count) is not int:
        error_msg = "count must be an integer"
        logger.error(error_msg)
        raise ValueError(error_msg)
    elif count < 0:
        error_msg = "count must be equal or greater than 0"
        logger.error(error_msg)
        raise ValueError(error_msg)

    if isinstance(repository, list):
        repository = ",".join(repository)
    if not isinstance(repository, str):
        error_msg = "repository must be a string or a list of strings"
        raise ValueError(error_msg)

    # build the payload
    app_key = get_default_session().app_key
    payload = {
        "number": str(count),
        "query": query,
        "repository": repository,
        "productName": app_key,
        "attributionCode": "",
    }

    if date_from is not None:
        payload.update({"dateFrom": to_datetime(date_from).isoformat()})

    if date_to is not None:
        payload.update({"dateTo": to_datetime(date_to).isoformat()})

    response = json_requests.send_json_request(
        News_Headlines_UDF_endpoint, payload, debug=debug
    )
    result = response.json()

    if raw_output:
        return result
    else:
        return convert_content_data_to_df_udf(result)


def get_news_story(story_id, raw_output=False, debug=False):
    """
    Return a single news story corresponding to the identifier provided in story_id

    Parameters
    ----------
    story_id:  The story id. The story id is a field you will find in every headline you retrieved
        with the legacy get_news_headlines

    raw_output: boolean
        Set this parameter to True to get the data in json format
        if set to False, the legacy will return returns the story content
        The default value is False

    debug: bool
        When set to True, the json request and response are printed.

    Raises
    ------
    Exception
        If http request fails or if Refinitiv Services return an error
    ValueError
        If a parameter type or value is wrong

    Examples
    --------
    >> import refinitiv.data as ek
    >> ek.set_app_key('set your app key here')
    >> headlines = ek.get_news_headlines('IBM')
    >> for index, headline_row in headlines.iterrows():
           story = ek.get_news_story(headline_row['storyId'])
           print (story)
    """
    logger = get_default_session().logger()

    # check parameters type and values
    if not is_string_type(story_id):
        error_msg = "story_id must be a string"
        logger.error(error_msg)
        raise ValueError(error_msg)

    app_key = get_default_session().app_key
    payload = {"attributionCode": "", "productName": app_key, "storyId": story_id}
    response = json_requests.send_json_request(
        News_Story_UDF_endpoint, payload, debug=debug
    )
    json_data = response.json()
    if raw_output:
        return json_data

    if json_data:
        if json_data.get("story"):
            if json_data.get("story").get("storyHtml"):
                return json_data.get("story").get("storyHtml")
            else:
                return None
        elif json_data.get("viewURL"):
            return json_data.get("viewURL")
        elif json_data.get("webURL"):
            return json_data.get("webURL")

    return None



# coding: utf8

__all__ = ["CallPut"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class CallPut(Enum):
    """
    Tells if the option is a call or a put.

    Possible values:
        - Call
        - Put
        - None
    """

    CALL = "CALL"
    NONE = "None"
    PUT = "PUT"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(CallPut, _CALL_PUT_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_CALL_PUT_VALUES_IN_LOWER_BY_CALL_PUT, some)


_CALL_PUT_VALUES = tuple(t.value for t in CallPut)
_CALL_PUT_VALUES_IN_LOWER_BY_CALL_PUT = {
    name.lower(): item for name, item in CallPut.__members__.items()
}



# coding: utf8


class StreamCacheIterator:
    """StreamCache Iterator class"""

    def __init__(self, stream_cache):
        if stream_cache.get_fields():
            self._field_values = list(stream_cache.get_fields().items())
        else:
            self._field_values = []
        self._index = 0

    def __next__(self):
        """Return the next field value from stream cache"""
        if self._index < len(self._field_values):
            result = self._field_values[self._index]
            self._index += 1
            return result
        raise StopIteration()


class StreamCache:
    """
    StreamCache contains all retrieved data from stream subscription.

    Raises
    ------
    Exception
        If request fails or if Refinitiv Services return an error
    """

    def __init__(self, name, fields=None, service=None, status=None, record=None):
        self._name = name
        self._fields = fields if fields else []
        self._service = service
        self._status = status
        self._record = record

    ###################################################
    #  Access to StreamCache as a dict                #
    ###################################################

    def keys(self):
        if self._record and self._record.get("Fields"):
            return self._record["Fields"].keys()
        return {}.keys()

    def values(self):
        if self._record and self._record.get("Fields"):
            return self._record["Fields"].values()
        return {}.values()

    def items(self):
        if self._record and self._record.get("Fields"):
            return self._record["Fields"].items()
        return {}.items()

    ###################################################
    #  Make StreamCache iterable                      #
    ###################################################

    def __iter__(self):
        return StreamCacheIterator(self)

    def __getitem__(self, field):
        if (
            self._record
            and self._record.get("Fields")
            and field in self._record["Fields"].keys()
        ):
            return self._record["Fields"][field]
        raise KeyError(f"Field '{field}' not in Stream cache")

    def __len__(self):
        return len(self._fields)

    def __repr__(self):
        rep = str(
            {
                "name": self.name,
                "service": self.service,
                "fields": dict(list(self.items())),
            }
        )
        return rep

    def __str__(self):
        if self.service:
            return str(
                "|".join([self.service, self.name])
                + "["
                + ",".join([f"{f}:{v}" for f, v in self.items()])
                + "]"
            )
        else:
            return str(
                "|".join(["Unknown service", self.name])
                + "["
                + ",".join([f"{f}:{v}" for f, v in self.items()])
                + "]"
            )

    ###################################################
    #  StreamCache properties                         #
    ###################################################

    @property
    def name(self):
        return self._name

    @property
    def service(self):
        return self._service

    @property
    def fields(self):
        if self._record and self._record.get("Fields"):
            return self._record["Fields"].keys()
        return None

    @property
    def status(self):
        return self._status

    @property
    def is_ok(self):
        return True if self._status.get("Data") == "Ok" else False

    # ###################################################
    # #  StreamingCache data accessors                  #
    # ###################################################

    def get_field_value(self, field):
        if (
            self._record
            and self._record.get("Fields")
            and field in self._record["Fields"].keys()
        ):
            return self._record["Fields"][field]

    def get_fields(self, fields=None) -> dict:
        if not self._record:
            return dict.fromkeys(fields or self._fields)

        if not fields:
            return self._record.get("Fields")

        return {**dict.fromkeys(fields), **self._record.get("Fields")}
